import asyncio
from typing import Optional
from unittest.mock import Mock, call

import pytest

from supriya import AddAction, CalculationRate
from supriya.assets.synthdefs import default, system_link_audio_1
from supriya.clocks import AsyncOfflineClock, OfflineClock
from supriya.contexts import AsyncServer, BusGroup, Context, Group, Score, Server, Synth
from supriya.contexts.requests import NewGroup
from supriya.osc import OscBundle, OscMessage
from supriya.patterns import (
    BusPattern,
    EventPattern,
    GroupPattern,
    MonoEventPattern,
    ParallelPattern,
    SequencePattern,
)
from supriya.patterns.events import NoteEvent, Priority, StartEvent, StopEvent


@pytest.mark.parametrize(
    "pattern, until, target_node, expected",
    [
        (
            SequencePattern(
                [
                    EventPattern(frequency=SequencePattern([440, 550, 660])),
                    MonoEventPattern(frequency=SequencePattern([440, 550, 660])),
                ]
            ),
            None,
            None,
            lambda context: [
                call.at(0.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=440,
                ),
                call.at(2.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=550,
                ),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1000,
                        synthdef=default,
                    )
                ),
                call.at(4.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=660,
                ),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1001,
                        synthdef=default,
                    )
                ),
                call.at(6.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=440,
                ),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1002,
                        synthdef=default,
                    )
                ),
                call.at(8.0),
                call.set_node(
                    Synth(
                        context=context,
                        id_=1003,
                        synthdef=default,
                    ),
                    frequency=550,
                ),
                call.at(10.0),
                call.set_node(
                    Synth(
                        context=context,
                        id_=1003,
                        synthdef=default,
                    ),
                    frequency=660,
                ),
                call.at(12.0),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1003,
                        synthdef=default,
                    )
                ),
            ],
        ),
        (
            EventPattern(frequency=SequencePattern([440, 550, 660])),
            1.5,
            None,
            lambda context: [
                call.at(0.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=440,
                ),
                call.at(2.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=550,
                ),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1000,
                        synthdef=default,
                    )
                ),
                call.at(3.0),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1001,
                        synthdef=default,
                    )
                ),
                call.at(3.0),  # 1001 was freed early, nothing to do.
            ],
        ),
        (
            EventPattern(frequency=SequencePattern([440, 550, 660])),
            1.5,
            6666,
            lambda context: [
                call.at(0.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=Group(context=context, id_=6666),
                    frequency=440,
                ),
                call.at(2.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=Group(context=context, id_=6666),
                    frequency=550,
                ),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1000,
                        synthdef=default,
                    )
                ),
                call.at(3.0),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1001,
                        synthdef=default,
                    )
                ),
                call.at(3.0),  # 1001 was freed early, nothing to do.
            ],
        ),
        (
            GroupPattern(EventPattern(frequency=SequencePattern([440, 550, 660]))),
            1.5,
            6666,
            lambda context: [
                call.at(0.0),
                call.add_group(
                    add_action=AddAction.ADD_TO_HEAD,
                    target_node=Group(context=context, id_=6666),
                ),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=Group(context=context, id_=1000),
                    frequency=440,
                ),
                call.at(2.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=Group(context=context, id_=1000),
                    frequency=550,
                ),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1001,
                        synthdef=default,
                    )
                ),
                call.at(3.0),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1002,
                        synthdef=default,
                    )
                ),
                call.at(3.0),  # 1001 was freed early, nothing to do.
                call.at(3.5),
                call.free_node(Group(context=context, id_=1000)),
            ],
        ),
        (
            ParallelPattern(
                [
                    EventPattern(frequency=SequencePattern([440, 550])),
                    EventPattern(frequency=SequencePattern([777, 888])),
                ]
            ),
            None,
            None,
            lambda context: [
                call.at(0.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=440,
                ),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=777,
                ),
                call.at(2.0),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=550,
                ),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=None,
                    frequency=888,
                ),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1000,
                        synthdef=default,
                    )
                ),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1001,
                        synthdef=default,
                    )
                ),
                call.at(4.0),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1002,
                        synthdef=default,
                    )
                ),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1003,
                        synthdef=default,
                    )
                ),
            ],
        ),
        (
            GroupPattern(
                BusPattern(MonoEventPattern(frequency=SequencePattern([440, 550, 660])))
            ),
            1.5,
            None,
            lambda context: [
                call.at(0.0),
                call.add_group(add_action=AddAction.ADD_TO_HEAD, target_node=None),
                call.add_bus_group(calculation_rate=CalculationRate.AUDIO, count=1),
                call.add_group(
                    add_action=AddAction.ADD_TO_HEAD,
                    target_node=Group(context=context, id_=1000),
                ),
                call.add_synth(
                    add_action=AddAction.ADD_AFTER,
                    permanent=False,
                    synthdef=system_link_audio_1,
                    target_node=Group(context=context, id_=1001),
                    amplitude=1.0,
                    fade_time=0.25,
                    in_=BusGroup(
                        context=context,
                        calculation_rate=CalculationRate.AUDIO,
                        id_=16,
                        count=1,
                    ),
                ),
                call.add_synth(
                    add_action=AddAction.ADD_TO_HEAD,
                    permanent=False,
                    synthdef=default,
                    target_node=Group(context=context, id_=1001),
                    frequency=440,
                    out=BusGroup(
                        context=context,
                        calculation_rate=CalculationRate.AUDIO,
                        id_=16,
                        count=1,
                    ),
                ),
                call.at(2.0),
                call.set_node(
                    Synth(
                        context=context,
                        id_=1003,
                        synthdef=default,
                    ),
                    frequency=550,
                    out=BusGroup(
                        context=context,
                        calculation_rate=CalculationRate.AUDIO,
                        id_=16,
                        count=1,
                    ),
                ),
                call.at(3.0),
                call.free_node(
                    Synth(
                        context=context,
                        id_=1003,
                        synthdef=default,
                    )
                ),
                call.at(3.0),  # Can we coalesce these moments?
                call.free_node(
                    Synth(
                        context=context,
                        id_=1002,
                        synthdef=system_link_audio_1,
                    )
                ),
                call.at(3.5),
                call.free_node(Group(context=context, id_=1001)),
                call.free_bus_group(
                    BusGroup(
                        context=context,
                        calculation_rate=CalculationRate.AUDIO,
                        id_=16,
                        count=1,
                    )
                ),
                call.free_node(Group(context=context, id_=1000)),
            ],
        ),
    ],
)
def test_context_calls(
    pattern, until: Optional[float], target_node: Optional[int], expected
):
    clock = OfflineClock()
    context = Server().boot()
    target_node_ = None
    if target_node is not None:
        context.send(
            NewGroup(items=[(target_node, "add_to_tail", 1)])
        )  # create an arbitrarily-ID'd group
        target_node_ = Group(context=context, id_=target_node)
    spy = Mock(spec=Context, wraps=context)
    pattern.play(context=spy, clock=clock, target_node=target_node_, until=until)
    expected_mock_calls = expected(context)
    assert spy.mock_calls == expected_mock_calls


def test_callback():
    def callback(player, context, event, priority):
        callback_calls.append(
            (player, context.desired_moment.offset, type(event), priority)
        )

    callback_calls = []
    pattern = EventPattern(frequency=SequencePattern([440, 550, 660]))
    clock = OfflineClock()
    player = pattern.play(context=Server().boot(), clock=clock, callback=callback)
    assert callback_calls == [
        (player, 0.0, StartEvent, Priority.START),
        (player, 0.0, NoteEvent, Priority.START),
        (player, 1.0, NoteEvent, Priority.START),
        (player, 1.0, NoteEvent, Priority.STOP),
        (player, 2.0, NoteEvent, Priority.START),
        (player, 2.0, NoteEvent, Priority.STOP),
        (player, 3.0, NoteEvent, Priority.STOP),
        (player, 3.0, StopEvent, Priority.STOP),
    ]


@pytest.mark.asyncio
async def test_callback_async():
    def callback(player, context, event, priority):
        print("CALLBACK", player, context, event, priority)
        callback_calls.append(
            (player, context.desired_moment.offset, type(event), priority)
        )
        if isinstance(event, StopEvent):
            stop_future.set_result(True)

    event_loop = asyncio.get_running_loop()
    stop_future = event_loop.create_future()
    callback_calls = []
    pattern = EventPattern(frequency=SequencePattern([440, 550, 660]))
    clock = AsyncOfflineClock()
    player = pattern.play(
        context=await AsyncServer().boot(), clock=clock, callback=callback
    )
    await clock.start()
    await stop_future
    await clock.stop()
    assert callback_calls == [
        (player, 0.0, StartEvent, Priority.START),
        (player, 0.0, NoteEvent, Priority.START),
        (player, 1.0, NoteEvent, Priority.START),
        (player, 1.0, NoteEvent, Priority.STOP),
        (player, 2.0, NoteEvent, Priority.START),
        (player, 2.0, NoteEvent, Priority.STOP),
        (player, 3.0, NoteEvent, Priority.STOP),
        (player, 3.0, StopEvent, Priority.STOP),
    ]


def test_nonrealtime():
    at = 1.0
    until = 1.5
    pattern = GroupPattern(
        BusPattern(MonoEventPattern(frequency=SequencePattern([440, 550, 660])))
    )
    clock = OfflineClock()
    context = Score()
    pattern.play(context=context, clock=clock, at=at, until=until)
    # Session should not map in_ or out, but use their bus numbers as consts.
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/g_new", 1000, 0, 0, 1001, 0, 1000),
                OscMessage(
                    "/s_new",
                    "system_link_audio_1",
                    1002,
                    3,
                    1001,
                    "fade_time",
                    0.25,
                    "in_",
                    16.0,
                ),
                OscMessage("/s_new", "default", 1003, 0, 1001, "out", 16.0),
            ),
            timestamp=1.0,
        ),
        OscBundle(
            contents=(OscMessage("/n_set", 1003, "frequency", 550.0, "out", 16.0),),
            timestamp=3.0,
        ),
        OscBundle(
            contents=(
                OscMessage("/n_set", 1003, "gate", 0.0),
                OscMessage("/n_set", 1002, "gate", 0.0),
            ),
            timestamp=4.0,
        ),
        OscBundle(
            contents=(
                OscMessage("/n_free", 1001),
                OscMessage("/n_free", 1000),
            ),
            timestamp=4.5,
        ),
    ]



from ..enums import SignalRange
from .core import UGen, param, ugen


@ugen(kr=True, signal_range=SignalRange.UNIPOLAR)
class KeyState(UGen):
    keycode = param(0.0)
    minimum = param(0.0)
    maximum = param(1.0)
    lag = param(0.2)


@ugen(kr=True, signal_range=SignalRange.UNIPOLAR)
class MouseButton(UGen):
    """
    A mouse-button tracker.

    ::

        >>> supriya.ugens.MouseButton.kr()
        <MouseButton.kr()[0]>
    """

    minimum = param(0.0)
    maximum = param(1.0)
    lag = param(0.2)


@ugen(kr=True, signal_range=SignalRange.UNIPOLAR)
class MouseX(UGen):
    """
    A mouse cursor tracker.

    MouseX tracks the y-axis of the mouse cursor position.

    ::

        >>> supriya.ugens.MouseX.kr()
        <MouseX.kr()[0]>
    """

    minimum = param(0.0)
    maximum = param(1.0)
    warp = param(0.0)
    lag = param(0.2)


@ugen(kr=True, signal_range=SignalRange.UNIPOLAR)
class MouseY(UGen):
    """
    A mouse cursor tracker.

    MouseY tracks the y-axis of the mouse cursor position.

    ::

        >>> supriya.ugens.MouseY.kr()
        <MouseY.kr()[0]>
    """

    minimum = param(0.0)
    maximum = param(1.0)
    warp = param(0.0)
    lag = param(0.2)



import contextlib
import dataclasses
import logging
import socket
import time
from collections.abc import Sequence as SequenceABC
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from ..enums import BootStatus
from ..typing import FutureLike
from .messages import OscBundle, OscMessage

osc_protocol_logger = logging.getLogger(__name__)
osc_in_logger = logging.getLogger("supriya.osc.in")
osc_out_logger = logging.getLogger("supriya.osc.out")
udp_in_logger = logging.getLogger("supriya.udp.in")
udp_out_logger = logging.getLogger("supriya.udp.out")


def find_free_port():
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


class OscProtocolOffline(Exception):
    pass


class OscProtocolAlreadyConnected(Exception):
    pass


class OscCallback(NamedTuple):
    pattern: Tuple[Union[str, int, float], ...]
    procedure: Callable
    failure_pattern: Optional[Tuple[Union[str, int, float], ...]] = None
    once: bool = False
    args: Optional[Tuple] = None
    kwargs: Optional[Dict] = None


@dataclasses.dataclass
class HealthCheck:
    request_pattern: List[str]
    response_pattern: List[str]
    active: bool = True
    timeout: float = 1.0
    backoff_factor: float = 1.5
    max_attempts: int = 5


class CaptureEntry(NamedTuple):
    timestamp: float
    label: str
    message: Union[OscMessage, OscBundle]


class Capture:
    ### INITIALIZER ###

    def __init__(self, osc_protocol):
        self.osc_protocol = osc_protocol
        self.messages = []

    ### SPECIAL METHODS ###

    def __enter__(self):
        self.osc_protocol.captures.add(self)
        self.messages[:] = []
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.osc_protocol.captures.remove(self)

    def __iter__(self):
        return iter(self.messages)

    def __len__(self):
        return len(self.messages)

    ### PUBLIC METHODS ###

    def filtered(
        self, sent=True, received=True, status=True
    ) -> List[Union[OscBundle, OscMessage]]:
        messages = []
        for _, label, message in self.messages:
            if label == "R" and not received:
                continue
            if label == "S" and not sent:
                continue
            if (
                isinstance(message, OscMessage)
                and message.address in ("/status", "/status.reply")
                and not status
            ):
                continue
            messages.append(message)
        return messages

    ### PUBLIC PROPERTIES ###

    @property
    def received_messages(self):
        return [
            (timestamp, osc_message)
            for timestamp, label, osc_message in self.messages
            if label == "R"
        ]

    @property
    def sent_messages(self):
        return [
            (timestamp, osc_message)
            for timestamp, label, osc_message in self.messages
            if label == "S"
        ]


class OscProtocol:
    ### INITIALIZER ###

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        on_connect_callback: Optional[Callable] = None,
        on_disconnect_callback: Optional[Callable] = None,
        on_panic_callback: Optional[Callable] = None,
    ) -> None:
        self.callbacks: Dict[Any, Any] = {}
        self.captures: Set[Capture] = set()
        self.healthcheck: Optional[HealthCheck] = None
        self.healthcheck_osc_callback: Optional[OscCallback] = None
        self.attempts = 0
        self.ip_address = "127.0.0.1"
        self.name = name
        self.port = 57551
        self.on_connect_callback = on_connect_callback
        self.on_disconnect_callback = on_disconnect_callback
        self.on_panic_callback = on_panic_callback
        self.status = BootStatus.OFFLINE

    ### PRIVATE METHODS ###

    def _activate_healthcheck(self) -> bool:
        if not self.healthcheck:
            return False
        elif self.healthcheck.active:
            return False
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "activating healthcheck..."
        )
        return True

    def _add_callback(self, callback: OscCallback) -> None:
        patterns = [callback.pattern]
        if callback.failure_pattern:
            patterns.append(callback.failure_pattern)
        for pattern in patterns:
            callback_map = self.callbacks
            for item in pattern:
                callbacks, callback_map = callback_map.setdefault(item, ([], {}))
            callbacks.append(callback)

    def _disconnect(self, panicked: bool = False) -> Optional[Awaitable[None]]:
        if panicked:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "panicking ..."
            )
        else:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "disconnecting ..."
            )
        self.status = BootStatus.QUITTING
        if self.healthcheck_osc_callback is not None:
            self.unregister(self.healthcheck_osc_callback)
        return None

    def _match_callbacks(self, message) -> List[OscCallback]:
        items = (message.address,) + message.contents
        matching_callbacks = []
        callback_map = self.callbacks
        for item in items:
            if item not in callback_map:
                break
            callbacks, callback_map = callback_map[item]
            matching_callbacks.extend(callbacks)
        for callback in matching_callbacks:
            if callback.once:
                self.unregister(callback)
        return matching_callbacks

    def _on_connect(
        self, *, boot_future: FutureLike[bool]
    ) -> Optional[Awaitable[None]]:
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "... connected!"
        )
        self.status = BootStatus.ONLINE
        boot_future.set_result(True)
        return None

    def _on_disconnect(
        self,
        *,
        boot_future: FutureLike[bool],
        exit_future: FutureLike[bool],
        panicked: bool = False,
    ) -> Optional[Awaitable[None]]:
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "... disconnected!"
        )
        self.status = BootStatus.OFFLINE
        if not boot_future.done():
            boot_future.set_result(False)
        if not exit_future.done():
            exit_future.set_result(not panicked)
        return None

    def _on_healthcheck_passed(self, message: OscMessage) -> Optional[Awaitable[None]]:
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "healthcheck: passed"
        )
        self.attempts = 0
        return None

    def _remove_callback(self, callback: OscCallback) -> None:
        def delete(pattern, original_callback_map):
            key = pattern.pop(0)
            if key not in original_callback_map:
                return
            callbacks, callback_map = original_callback_map[key]
            if pattern:
                delete(pattern, callback_map)
            if callback in callbacks:
                callbacks.remove(callback)
            if not callbacks and not callback_map:
                original_callback_map.pop(key)

        patterns = [callback.pattern]
        if callback.failure_pattern:
            patterns.append(callback.failure_pattern)
        for pattern in patterns:
            delete(list(pattern), self.callbacks)

    def _register(
        self,
        pattern,
        procedure,
        *,
        failure_pattern=None,
        once: bool = False,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict] = None,
    ) -> OscCallback:
        if isinstance(pattern, (str, int, float)):
            pattern = [pattern]
        if isinstance(failure_pattern, (str, int, float)):
            failure_pattern = [failure_pattern]
        if not callable(procedure):
            raise ValueError(procedure)
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            f"registering pattern: {pattern!r}"
        )
        return OscCallback(
            pattern=tuple(pattern),
            failure_pattern=failure_pattern,
            procedure=procedure,
            once=bool(once),
            args=args,
            kwargs=kwargs,
        )

    def _send(self, message):
        if self.status not in (BootStatus.BOOTING, BootStatus.ONLINE):
            raise OscProtocolOffline
        if not isinstance(message, (str, SequenceABC, OscBundle, OscMessage)):
            raise ValueError(message)
        if isinstance(message, str):
            message = OscMessage(message)
        elif isinstance(message, SequenceABC):
            message = OscMessage(*message)
        osc_out_logger.debug(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            f"{message!r}"
        )
        for capture in self.captures:
            capture.messages.append(
                CaptureEntry(timestamp=time.time(), label="S", message=message)
            )
        datagram = message.to_datagram()
        udp_out_logger.debug(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            f"{datagram}"
        )
        return datagram

    def _setup(
        self, ip_address: str, port: int, healthcheck: Optional[HealthCheck]
    ) -> None:
        self.status = BootStatus.BOOTING
        self.ip_address = ip_address
        self.port = port
        self.healthcheck = healthcheck
        if self.healthcheck:
            self.healthcheck_osc_callback = self.register(
                pattern=self.healthcheck.response_pattern,
                procedure=self._on_healthcheck_passed,
            )

    def _validate_receive(self, datagram):
        udp_in_logger.debug(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            f"{datagram}"
        )
        try:
            message = OscMessage.from_datagram(datagram)
        except Exception:
            raise
        osc_in_logger.debug(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            f"{message!r}"
        )
        for capture in self.captures:
            capture.messages.append(
                CaptureEntry(timestamp=time.time(), label="R", message=message)
            )
        for callback in self._match_callbacks(message):
            yield callback, message

    ### PUBLIC METHODS ###

    def activate_healthcheck(self) -> None:
        raise NotImplementedError

    def capture(self) -> "Capture":
        return Capture(self)

    def disconnect(self) -> Optional[Awaitable[None]]:
        raise NotImplementedError

    def register(
        self,
        pattern: Sequence[Union[str, float]],
        procedure: Callable[[OscMessage], Optional[Awaitable[None]]],
        *,
        failure_pattern: Optional[Sequence[Union[str, float]]] = None,
        once: bool = False,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict] = None,
    ) -> OscCallback:
        raise NotImplementedError

    def send(self, message: Union[OscBundle, OscMessage, SequenceABC, str]) -> None:
        raise NotImplementedError

    def unregister(self, callback: OscCallback) -> None:
        raise NotImplementedError



import operator

import pytest

from supriya.patterns import BinaryOpPattern, SequencePattern
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, operator, input_a, input_b, expected, is_infinite",
    [
        (None, operator.add, 1, 7, [8], True),
        (None, operator.add, [1], 7, [(8,)], True),
        (None, operator.add, [1], [7], [(8,)], True),
        (None, operator.add, [[1]], [7], [((8,),)], True),
        (None, operator.add, [[[1]]], [7], [(((8,),),)], True),
        (None, operator.add, [1, 2], [7, 8], [(8, 10)], True),
        (None, operator.add, [1, 2], [7, 8, 10], [(8, 10, 11)], True),
        (None, operator.mul, SequencePattern([1, 2, 3]), 4, [4, 8, 12], False),
        (None, operator.mul, SequencePattern([1, 2, 3], None), 4, [4, 8, 12], True),
        (
            None,
            operator.add,
            SequencePattern(["a", "b", "c"]),
            SequencePattern(["1", "2", "3"]),
            ["a1", "b2", "c3"],
            False,
        ),
    ],
)
def test(stop_at, operator, input_a, input_b, expected, is_infinite):
    pattern = BinaryOpPattern(operator, input_a, input_b)
    run_pattern_test(pattern, expected, is_infinite, stop_at)



import operator

import pytest

from supriya.patterns import SequencePattern, UnaryOpPattern
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, operator, input_, expected, is_infinite",
    [
        (None, operator.neg, 1, [-1], True),
        (None, operator.neg, [1], [(-1,)], True),
        (None, operator.neg, [[1]], [((-1,),)], True),
        (None, operator.neg, [[[1]]], [(((-1,),),)], True),
        (None, operator.neg, [1, 2], [(-1, -2)], True),
        (None, operator.neg, SequencePattern([1, 2, 3]), [-1, -2, -3], False),
        (None, operator.neg, SequencePattern([1, 2, 3], None), [-1, -2, -3], True),
    ],
)
def test(stop_at, operator, input_, expected, is_infinite):
    pattern = UnaryOpPattern(operator, input_)
    run_pattern_test(pattern, expected, is_infinite, stop_at)



import pytest

from supriya.patterns import RandomPattern


@pytest.mark.parametrize(
    "minimum, maximum, iterations, is_infinite",
    [
        (0.0, 1.0, None, True),
        (0.0, 1.0, 1, False),
        (0.45, 0.55, None, True),
        (0.0, (1.0, 2.0), None, True),
    ],
)
def test(minimum, maximum, iterations, is_infinite):
    pattern = RandomPattern(minimum=minimum, maximum=maximum, iterations=iterations)
    assert pattern.distribution == RandomPattern.Distribution.WHITE_NOISE
    assert pattern.is_infinite == is_infinite
    assert pattern.iterations == iterations
    assert pattern.maximum == maximum
    assert pattern.minimum == minimum
    iterator = iter(pattern)
    ceased = True
    actual = []
    for _ in range(1000):
        try:
            actual.append(next(iterator))
        except StopIteration:
            break
    else:
        ceased = False
    if is_infinite:
        assert not ceased
    else:
        assert len(actual) == iterations
    # TODO: Verify minimum / maximum bounds
    assert len(set(actual)) == len(actual)



import enum
from typing import Any, Dict, List, Optional, Sequence, SupportsFloat, Tuple, Union
from uuid import UUID

from uqbar.objects import get_repr, get_vars, new

from ..assets.synthdefs.default import default
from ..contexts import BusGroup, Context, ContextObject, Node
from ..enums import AddAction, CalculationRate
from ..typing import AddActionLike, CalculationRateLike
from ..ugens import SynthDef
from ..utils import expand


class Priority(enum.IntEnum):
    NONE = 0
    START = 1
    STOP = 2


class Event:
    def __init__(self, *, delta: float = 0.0, **kwargs) -> None:
        self.delta = delta

    def __eq__(self, expr: Any) -> bool:
        self_values = type(self), get_vars(self)
        try:
            expr_values = type(expr), get_vars(expr)
        except AttributeError:
            expr_values = type(expr), expr
        return self_values == expr_values

    def __repr__(self) -> str:
        return get_repr(self, multiline=False)

    def expand(self, offset: float) -> Sequence[Tuple[float, Priority, "Event"]]:
        return [(offset, Priority.START, self)]

    def perform(
        self,
        context: Context,
        proxy_mapping: Dict[Union[UUID, Tuple[UUID, int]], ContextObject],
        *,
        current_offset: float,
        notes_mapping: Dict[Union[UUID, Tuple[UUID, int]], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        pass


class StartEvent(Event):
    """
    The first event injected by a pattern player.
    """


class StopEvent(Event):
    """
    The last event injected by a pattern player.
    """


class BusAllocateEvent(Event):
    def __init__(
        self,
        id_: Union[UUID, Tuple[UUID, int]],
        *,
        calculation_rate: CalculationRateLike = "audio",
        channel_count: int = 1,
        delta: float = 0.0,
        **kwargs,
    ) -> None:
        Event.__init__(self, delta=delta, **kwargs)
        self.id_ = id_
        self.calculation_rate = CalculationRate.from_expr(calculation_rate)
        self.channel_count = channel_count

    def perform(
        self,
        context: Context,
        proxy_mapping: Dict[Union[UUID, Tuple[UUID, int]], ContextObject],
        *,
        current_offset: float,
        notes_mapping: Dict[Union[UUID, Tuple[UUID, int]], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        proxy_mapping[self.id_] = context.add_bus_group(
            calculation_rate=self.calculation_rate, count=self.channel_count
        )


class BusFreeEvent(Event):
    def __init__(
        self, id_: Union[UUID, Tuple[UUID, int]], *, delta: float = 0.0, **kwargs
    ) -> None:
        Event.__init__(self, delta=delta, **kwargs)
        self.id_ = id_

    def perform(
        self,
        context: Context,
        proxy_mapping: Dict[Union[UUID, Tuple[UUID, int]], ContextObject],
        *,
        current_offset: float,
        notes_mapping: Dict[Union[UUID, Tuple[UUID, int]], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        if not isinstance(bus_group := proxy_mapping.pop(self.id_), BusGroup):
            raise RuntimeError(bus_group)
        context.free_bus_group(bus_group)


class CompositeEvent(Event):
    def __init__(self, events: List[Event], *, delta: float = 0.0, **kwargs) -> None:
        Event.__init__(self, delta=delta)
        self.events = (
            events if not kwargs else [new(event, **kwargs) for event in events]
        )

    def expand(self, offset: float) -> Sequence[Tuple[float, Priority, Event]]:
        events: List[Tuple[float, Priority, Event]] = []
        for event in self.events:
            events.extend(event.expand(offset))
            offset += event.delta
        return events


class NodeEvent(Event):
    def __init__(
        self,
        id_: Union[UUID, Tuple[UUID, int]],
        *,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        delta: float = 0.0,
        target_node: Optional[Union[Node, UUID]] = None,
        **kwargs,
    ) -> None:
        Event.__init__(self, delta=delta, **kwargs)
        self.id_ = id_
        self.add_action = AddAction.from_expr(add_action)
        self.target_node = target_node

    def _resolve_target_node(
        self,
        proxy_mapping: Dict[Union[UUID, Tuple[UUID, int]], ContextObject],
    ) -> Optional[Node]:
        if isinstance(self.target_node, UUID):
            if not isinstance(
                target_node_ := proxy_mapping.get(self.target_node), Node
            ):
                raise RuntimeError(target_node_)
            return target_node_
        return self.target_node


class GroupAllocateEvent(NodeEvent):
    def perform(
        self,
        context: Context,
        proxy_mapping: Dict[Union[UUID, Tuple[UUID, int]], ContextObject],
        *,
        current_offset: float,
        notes_mapping: Dict[Union[UUID, Tuple[UUID, int]], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        proxy_mapping[self.id_] = context.add_group(
            add_action=self.add_action,
            target_node=self._resolve_target_node(
                proxy_mapping,
            ),
        )


class NodeFreeEvent(Event):
    def __init__(
        self, id_: Union[UUID, Tuple[UUID, int]], *, delta: float = 0.0, **kwargs
    ) -> None:
        Event.__init__(self, delta=delta, **kwargs)
        self.id_ = id_

    def perform(
        self,
        context: Context,
        proxy_mapping: Dict[Union[UUID, Tuple[UUID, int]], ContextObject],
        *,
        current_offset: float,
        notes_mapping: Dict[Union[UUID, Tuple[UUID, int]], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        if not isinstance(node := proxy_mapping.pop(self.id_), Node):
            raise RuntimeError(node)
        context.free_node(node)


class NoteEvent(NodeEvent):
    def __init__(
        self,
        id_: Union[UUID, Tuple[UUID, int]],
        *,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        delta: float = 1.0,
        duration: float = 1.0,
        synthdef: Optional[SynthDef] = None,
        target_node: Optional[Union[Node, UUID]] = None,
        **kwargs: Union[
            Union[SupportsFloat, UUID],
            Sequence[Union[SupportsFloat, UUID]],
        ],
    ) -> None:
        NodeEvent.__init__(
            self,
            add_action=add_action,
            delta=delta,
            id_=id_,
            target_node=target_node,
            **kwargs,
        )
        self.duration = duration
        self.synthdef = synthdef
        self.kwargs = kwargs

    def expand(self, offset: float) -> Sequence[Tuple[float, Priority, Event]]:
        starts, stops = [], []
        for i, proxy_mapping in enumerate(expand(self.kwargs)):
            if not isinstance(self.id_, UUID):
                raise RuntimeError("How did we get here?")
            event: Event = type(self)(
                id_=(self.id_, i),
                add_action=self.add_action,
                delta=self.delta,
                duration=self.duration,
                synthdef=self.synthdef,
                target_node=self.target_node,
                **proxy_mapping,
            )
            start_offset = offset
            stop_offset = offset + self.duration
            starts.append((start_offset, Priority.START, event))
            stops.append((stop_offset, Priority.STOP, event))
        return starts + stops

    def merge(self, event: Event) -> Event:
        _, _, kwargs = get_vars(event)
        return new(self, **kwargs)

    def perform(
        self,
        context: Context,
        proxy_mapping: Dict[Union[UUID, Tuple[UUID, int]], ContextObject],
        *,
        current_offset: float,
        notes_mapping: Dict[Union[UUID, Tuple[UUID, int]], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        if priority == Priority.START:
            # does a proxy exist?
            #    if yes, update settings
            #    if no, create proxy
            # update notes mapping with expected completion offset
            settings: Dict[str, Union[SupportsFloat, Sequence[SupportsFloat]]] = {}
            for key, value in self.kwargs.items():
                if isinstance(value, UUID):
                    settings[key] = proxy_mapping[value]
                elif isinstance(value, Sequence):
                    settings[key] = [
                        float(proxy_mapping[x] if isinstance(x, UUID) else x)
                        for x in value
                    ]
                else:
                    settings[key] = float(value)
            # add the synth
            if self.id_ not in proxy_mapping:
                proxy_mapping[self.id_] = context.add_synth(
                    add_action=self.add_action,
                    permanent=False,
                    synthdef=self.synthdef or default,
                    target_node=self._resolve_target_node(
                        proxy_mapping,
                    ),
                    **settings,
                )
            else:
                if not isinstance(node := proxy_mapping.get(self.id_), Node):
                    raise RuntimeError(node)
                context.set_node(node, **settings)
            if self.duration:
                notes_mapping[self.id_] = current_offset + self.duration
        elif priority == Priority.STOP:
            # check notes mapping for expected completion offset:
            # if expected completion >= current_offset or non-existent? bail
            # otherwise release
            expected_completion = notes_mapping.get(self.id_)
            if expected_completion is None:
                proxy_mapping.pop(self.id_, None)
            elif expected_completion > current_offset:
                return
            else:
                notes_mapping.pop(self.id_)
                if not isinstance(node := proxy_mapping.pop(self.id_), Node):
                    raise RuntimeError(node)
                context.free_node(node)


class NullEvent(Event):
    def expand(self, offset: float) -> Sequence[Tuple[float, Priority, Event]]:
        return []


class SynthAllocateEvent(NodeEvent):
    def __init__(
        self,
        id_: Union[UUID, Tuple[UUID, int]],
        synthdef: SynthDef,
        *,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        delta: float = 0.0,
        target_node: Optional[Union[Node, UUID]] = None,
        **kwargs: Union[SupportsFloat, Sequence[SupportsFloat], UUID],
    ) -> None:
        NodeEvent.__init__(
            self,
            add_action=add_action,
            delta=delta,
            id_=id_,
            target_node=target_node,
        )
        self.synthdef = synthdef
        self.kwargs = kwargs

    def perform(
        self,
        context: Context,
        proxy_mapping: Dict[Union[UUID, Tuple[UUID, int]], ContextObject],
        *,
        current_offset: float,
        notes_mapping: Dict[Union[UUID, Tuple[UUID, int]], float],
        priority: Priority,
        **kwargs,
    ) -> None:
        settings: Dict[str, Union[SupportsFloat, Sequence[SupportsFloat]]] = {}
        for key, value in self.kwargs.items():
            if isinstance(value, UUID):
                value = proxy_mapping[value]
            settings[key] = value
        # add the synth
        proxy_mapping[self.id_] = context.add_synth(
            add_action=self.add_action,
            permanent=False,
            synthdef=self.synthdef,
            target_node=self._resolve_target_node(
                proxy_mapping,
            ),
            **settings,
        )



from queue import Empty, PriorityQueue, Queue
from threading import RLock
from typing import (
    Callable,
    Coroutine,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)
from uuid import UUID, uuid4
from weakref import WeakSet

from ..clocks import (
    BaseClock,
    CallbackEvent,
    Clock,
    ClockContext,
    OfflineClock,
    Quantization,
)
from ..contexts import Bus, Context, ContextObject, Node
from .eventpatterns import Pattern
from .events import Event, Priority, StartEvent, StopEvent
from .structure import PinPattern


class PatternPlayer:
    """
    A pattern player.

    Coordinates interactions between a pattern, a clock_context, and a clock.
    """

    # TODO: Rewrite type annotation after dropping 3.8
    _players = cast(Set["PatternPlayer"], WeakSet())

    def __init__(
        self,
        pattern: Pattern,
        context: Context,
        clock: BaseClock,
        callback: Optional[
            Callable[
                ["PatternPlayer", ClockContext, Event, Priority], Optional[Coroutine]
            ]
        ] = None,
        target_bus: Optional[Bus] = None,
        target_node: Optional[Node] = None,
        uuid: Optional[UUID] = None,
    ) -> None:
        self._context = context
        self._clock = clock
        self._callback = callback
        self._lock = RLock()
        self._queue: Queue[
            Tuple[float, Priority, Union[int, Tuple[int, int]], Optional[Event]]
        ] = PriorityQueue()
        self._is_running = False
        self._is_stopping = False
        self._proxies_by_uuid: Dict[Union[UUID, Tuple[UUID, int]], ContextObject] = {}
        self._notes_by_uuid: Dict[Union[UUID, Tuple[UUID, int]], float] = {}
        self._uuid: UUID = uuid or uuid4()
        self._target_bus = target_bus
        self._target_node = target_node
        self._next_delta: Optional[float] = None
        self._initial_seconds: Optional[float] = None
        self._pattern = (
            pattern
            if (target_bus is None and target_node is None)
            else PinPattern(pattern, target_bus=target_bus, target_node=target_node)
        )
        self._yielded = False

    def _clock_callback(
        self, clock_context: ClockContext, *args, **kwargs
    ) -> Optional[float]:
        for clock_context, seconds, offset, events in self._find_events(clock_context):
            if self._initial_seconds is None:
                self._initial_seconds = seconds
            with self._context.at(seconds):
                for event, priority in events:
                    event.perform(
                        self._context,
                        self._proxies_by_uuid,
                        current_offset=offset,
                        notes_mapping=self._notes_by_uuid,
                        priority=priority,
                    )
                    if self._callback is not None:
                        self._callback(self, clock_context, event, priority)
        return self._next_delta

    def _find_events(
        self, clock_context: ClockContext
    ) -> Iterable[Tuple[ClockContext, float, float, Sequence[Tuple[Event, Priority]]]]:
        with self._lock:
            current_offset = float("-inf")
            events: List[Tuple[Event, Priority]] = []
            if (
                not cast(CallbackEvent, clock_context.event).invocations
                and self._callback
            ):
                yield clock_context, clock_context.desired_moment.seconds, clock_context.desired_moment.offset, [
                    (StartEvent(), Priority.START)
                ]
            while True:
                try:
                    offset, priority, index, event = self._queue.get(block=False)
                except Empty:
                    if events:
                        yield clock_context, clock_context.desired_moment.seconds, current_offset, events
                    self._is_running = False
                    if self._callback:
                        yield clock_context, clock_context.desired_moment.seconds, current_offset, [
                            (StopEvent(), Priority.STOP)
                        ]
                    self._next_delta = None
                    return
                except Exception:
                    self._next_delta = None
                    return
                if offset == float("-inf"):
                    offset = clock_context.desired_moment.offset
                delta = offset - clock_context.desired_moment.offset
                if delta:
                    self._queue.put((offset, priority, index, event))
                    if events:
                        yield clock_context, clock_context.desired_moment.seconds, current_offset, events
                    self._next_delta = delta
                    return
                if not isinstance(event, Event):
                    if self._consume_iterator(offset):
                        if self._callback:
                            yield clock_context, clock_context.desired_moment.seconds, current_offset, [
                                (StopEvent(), Priority.STOP)
                            ]
                        self._next_delta = None
                        return
                elif offset != current_offset:
                    if events:
                        yield clock_context, clock_context.desired_moment.seconds, current_offset, events
                    current_offset = offset
                    events = [(event, priority)]
                else:
                    events.append((event, priority))

    def _consume_iterator(self, current_offset: float) -> bool:
        try:
            if self._yielded:
                index, consumed_event = self._iterator.send(self._is_stopping)
            else:
                if self._is_stopping:
                    return True
                index, consumed_event = next(self._iterator)
                self._yielded = True
            for subindex, (expanded_offset, priority, expanded_event) in enumerate(
                consumed_event.expand(current_offset)
            ):
                self._queue.put(
                    (expanded_offset, priority, (index, subindex), expanded_event)
                )
            self._queue.put(
                (
                    float(current_offset + consumed_event.delta),
                    Priority.NONE,
                    (index, 0),
                    None,
                )
            )
        except StopIteration:
            pass
        return False

    def _stop_callback(
        self, clock_context: ClockContext, *args, **kwargs
    ) -> Optional[float]:
        with self._lock:
            # Do we need to rebuild the queue? Yes.
            # Do we need to free all playing notes? Yes.
            # How do we handle when there are already stop events in the queue?
            # They'll be no-ops when performed.
            self._is_stopping = True
            self._clock.reschedule(
                self._clock_event_id, schedule_at=clock_context.desired_moment.offset
            )
            self._reschedule_queue(clock_context.desired_moment.offset)
            self._free_all_notes(clock_context.desired_moment.seconds)
        return None

    def _enumerate(
        self, iterator: Generator[Event, bool, None]
    ) -> Generator[Tuple[int, Event], bool, None]:
        index = 0
        should_stop = False
        while True:
            try:
                if index:
                    should_stop = (
                        yield (index, iterator.send(should_stop)) or should_stop
                    )
                else:
                    should_stop = yield (index, next(iterator))
            except StopIteration:
                return
            index += 1

    def _free_all_notes(self, current_seconds: float) -> None:
        if not self._notes_by_uuid:
            return
        with self._context.at(current_seconds):
            while self._notes_by_uuid:
                uuid, _ = self._notes_by_uuid.popitem()
                if not isinstance(node := self._proxies_by_uuid.pop(uuid), Node):
                    raise RuntimeError
                self._context.free_node(node)

    def _reschedule_queue(self, current_offset: float) -> None:
        events: List[
            Tuple[float, Priority, Union[int, Tuple[int, int]], Optional[Event]]
        ] = []
        while not self._queue.empty():
            events.append(self._queue.get())
        if not events:
            return
        delta = events[0][0] - current_offset
        for event in events:
            offset, priority, index, event_ = event
            self._queue.put((offset - delta, priority, index, event_))

    def play(
        self,
        quantization: Optional[Quantization] = None,
        at: Optional[float] = None,
        until: Optional[float] = None,
    ) -> None:
        with self._lock:
            if self._is_running:
                return
            self._iterator = self._enumerate(iter(self._pattern))
            self._queue.put((float("-inf"), Priority.NONE, (0, 0), None))
            self._is_running = True
            self._is_stopping = False
            self._yielded = False
            self._players.add(self)
        self._clock_event_id = self._clock.cue(
            self._clock_callback,
            event_type=3,
            quantization=quantization,
        )
        if until:
            self._clock.schedule(self._stop_callback, event_type=2, schedule_at=until)
        if (
            isinstance(self._clock, (Clock, OfflineClock))
            and not self._clock.is_running
        ):
            self._clock.start(initial_time=at)

    def stop(self, quantization: Optional[Quantization] = None) -> None:
        with self._lock:
            if not self._is_running or self._is_stopping:
                return
            self._clock.cue(
                self._stop_callback, event_type=2, quantization=quantization
            )
            self._players.remove(self)

    def uuid_to_note_id(self, uuid: UUID, index: Optional[int] = None) -> float:
        if index is not None:
            return self._notes_by_uuid[uuid, index]
        return self._notes_by_uuid[uuid]

    @property
    def initial_seconds(self) -> Optional[float]:
        return self._initial_seconds

    @property
    def uuid(self) -> UUID:
        return self._uuid



from typing import Any, Dict, Tuple

from uqbar.enums import IntEnumeration

from ..enums import CalculationRate
from .core import UGen, param, ugen


@ugen(kr=True, channel_count=4, fixed_channel_count=True)
class BeatTrack(UGen):
    """
    Autocorrelation beat tracker.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT.kr(source=source)
        >>> beat_track = supriya.ugens.BeatTrack.kr(
        ...     pv_chain=pv_chain,
        ...     lock=0,
        ... )
        >>> beat_track
        <BeatTrack.kr()>
    """

    pv_chain = param()
    lock = param(0.0)


@ugen(kr=True, channel_count=6, fixed_channel_count=True)
class BeatTrack2(UGen):
    """
    A template-matching beat-tracker.

    ::

        >>> beat_track_2 = supriya.ugens.BeatTrack2.kr(
        ...     bus_index=0,
        ...     lock=False,
        ...     feature_count=4,
        ...     phase_accuracy=0.02,
        ...     weighting_scheme=-2.1,
        ...     window_size=2,
        ... )
        >>> beat_track_2
        <BeatTrack2.kr()>
    """

    bus_index = param(0.0)
    feature_count = param()
    window_size = param(2)
    phase_accuracy = param(0.02)
    lock = param(0.0)
    weighting_scheme = param(-2.1)


@ugen(kr=True)
class KeyTrack(UGen):
    """
    A key tracker.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT.kr(source=source)
        >>> key_track = supriya.ugens.KeyTrack.kr(
        ...     pv_chain=pv_chain,
        ...     chroma_leak=0.5,
        ...     key_decay=2,
        ... )
        >>> key_track
        <KeyTrack.kr()[0]>
    """

    pv_chain = param()
    key_decay = param(2)
    chroma_leak = param(0.5)


@ugen(kr=True)
class Loudness(UGen):
    """
    Extraction of instantaneous loudness in `sones`.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT.kr(source=source)
        >>> loudness = supriya.ugens.Loudness.kr(
        ...     pv_chain=pv_chain,
        ...     smask=0.25,
        ...     tmask=1,
        ... )
        >>> loudness
        <Loudness.kr()[0]>
    """

    pv_chain = param()
    smask = param(0.25)
    tmask = param(1)


@ugen(kr=True, fixed_channel_count=True)
class MFCC(UGen):
    """
    Mel frequency cepstral coefficients.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT.kr(source=source)
        >>> mfcc = supriya.ugens.MFCC.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> mfcc
        <MFCC.kr()>
    """

    pv_chain = param()
    coeff_count = param(13)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs,
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        self._channel_count = int(kwargs["coeff_count"])
        return calculation_rate, kwargs


@ugen(kr=True)
class Onsets(UGen):
    """
    An onset detector.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT.kr(source=source)
        >>> onsets = supriya.ugens.Onsets.kr(
        ...     pv_chain=pv_chain,
        ...     floor_=0.1,
        ...     medianspan=11,
        ...     mingap=10,
        ...     odftype=supriya.ugens.Onsets.ODFType.RCOMPLEX,
        ...     rawodf=0,
        ...     relaxtime=1,
        ...     threshold=0.5,
        ...     whtype=1,
        ... )
        >>> onsets
        <Onsets.kr()[0]>
    """

    class ODFType(IntEnumeration):
        POWER = 0
        MAGSUM = 1
        COMPLEX = 2
        RCOMPLEX = 3
        PHASE = 4
        WPHASE = 5
        MKL = 6

    pv_chain = param()
    threshold = param(0.5)
    odftype = param(3)
    relaxtime = param(1)
    floor_ = param(0.1)
    mingap = param(10)
    medianspan = param(11)
    whtype = param(1)
    rawodf = param(0)


@ugen(kr=True, channel_count=2, fixed_channel_count=True)
class Pitch(UGen):
    """
    An autocorrelation pitch follower.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pitch = supriya.ugens.Pitch.kr(source=source)
        >>> pitch
        <Pitch.kr()>
    """

    source = param()
    initial_frequency = param(440)
    min_frequency = param(60)
    max_frequency = param(4000)
    exec_frequency = param(100)
    max_bins_per_octave = param(16)
    median = param(1)
    amplitude_threshold = param(0.01)
    peak_threshold = param(0.5)
    down_sample_factor = param(1)
    clarity = param(0)


@ugen(kr=True)
class SpecCentroid(UGen):
    """
    A spectral centroid measure.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT.kr(source=source)
        >>> spec_centroid = supriya.ugens.SpecCentroid.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> spec_centroid
        <SpecCentroid.kr()[0]>
    """

    pv_chain = param()


@ugen(kr=True)
class SpecFlatness(UGen):
    """
    A spectral flatness measure.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT.kr(source=source)
        >>> spec_flatness = supriya.ugens.SpecFlatness.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> spec_flatness
        <SpecFlatness.kr()[0]>
    """

    pv_chain = param()


@ugen(kr=True)
class SpecPcile(UGen):
    """
    Find a percentile of FFT magnitude spectrum.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT.kr(source=source)
        >>> spec_pcile = supriya.ugens.SpecPcile.kr(
        ...     pv_chain=pv_chain,
        ...     fraction=0.5,
        ...     interpolate=0,
        ... )
        >>> spec_pcile
        <SpecPcile.kr()[0]>
    """

    pv_chain = param()
    fraction = param(0.5)
    interpolate = param(0)



import concurrent.futures
import socketserver
import threading
import time
from collections.abc import Sequence as SequenceABC
from queue import Empty, Queue
from typing import (
    Awaitable,
    Callable,
    Dict,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from ..enums import BootStatus
from ..typing import FutureLike
from .messages import OscBundle, OscMessage
from .protocols import (
    HealthCheck,
    OscCallback,
    OscProtocol,
    OscProtocolAlreadyConnected,
    osc_protocol_logger,
)


class ThreadedOscProtocol(OscProtocol):

    class Server(socketserver.UDPServer):
        osc_protocol: "ThreadedOscProtocol"

        def verify_request(self, request, client_address) -> bool:
            self.osc_protocol._process_command_queue()
            return True

        def service_actions(self) -> None:
            if cast(HealthCheck, self.osc_protocol.healthcheck).active:
                self.osc_protocol._run_healthcheck()

    class Handler(socketserver.BaseRequestHandler):
        def handle(self) -> None:
            data = self.request[0]
            for callback, message in cast(
                ThreadedOscProtocol.Server, self.server
            ).osc_protocol._validate_receive(data):
                callback.procedure(
                    message, *(callback.args or ()), **(callback.kwargs or {})
                )

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        on_connect_callback: Optional[Callable] = None,
        on_disconnect_callback: Optional[Callable] = None,
        on_panic_callback: Optional[Callable] = None,
    ):
        OscProtocol.__init__(
            self,
            name=name,
            on_connect_callback=on_connect_callback,
            on_disconnect_callback=on_disconnect_callback,
            on_panic_callback=on_panic_callback,
        )
        self.boot_future: concurrent.futures.Future[bool] = concurrent.futures.Future()
        self.exit_future: concurrent.futures.Future[bool] = concurrent.futures.Future()
        self.command_queue: Queue[Tuple[Literal["add", "remove"], OscCallback]] = (
            Queue()
        )
        self.lock = threading.RLock()
        self.osc_server = self._server_factory(self.ip_address, self.port)
        self.osc_server_thread = threading.Thread(target=self.osc_server.serve_forever)

    ### PRIVATE METHODS ###

    def _disconnect(self, panicked: bool = False) -> None:
        super()._disconnect(panicked=panicked)
        if not self.osc_server._BaseServer__shutdown_request:
            # We set the shutdown request flag rather than call .shutdown()
            # because this is often being called from _inside_ the server
            # thread.
            self.osc_server._BaseServer__shutdown_request = True
        self._on_disconnect(
            boot_future=self.boot_future,
            exit_future=self.exit_future,
            panicked=panicked,
        )

    def _on_connect(self, boot_future: FutureLike[bool]) -> None:
        super()._on_connect(boot_future=boot_future)
        if self.on_connect_callback:
            self.on_connect_callback()

    def _on_disconnect(
        self,
        *,
        boot_future: FutureLike[bool],
        exit_future: FutureLike[bool],
        panicked: bool = False,
    ) -> None:
        super()._on_disconnect(
            boot_future=boot_future,
            exit_future=exit_future,
            panicked=panicked,
        )
        if panicked and self.on_panic_callback:
            self.on_panic_callback()
        elif not panicked and self.on_disconnect_callback:
            self.on_disconnect_callback()

    def _on_healthcheck_passed(self, message: OscMessage) -> None:
        super()._on_healthcheck_passed(message)
        if self.status == BootStatus.BOOTING:
            self._on_connect(boot_future=self.boot_future)

    def _process_command_queue(self):
        while self.command_queue.qsize():
            try:
                action, callback = self.command_queue.get()
            except Empty:
                continue
            if action == "add":
                self._add_callback(callback)
            elif action == "remove":
                self._remove_callback(callback)

    def _run_healthcheck(self):
        if self.healthcheck is None:
            return
        now = time.time()
        if now < self.healthcheck_deadline:
            return
        if self.attempts > 0:
            remaining = self.healthcheck.max_attempts - self.attempts
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                f"healthcheck failed, {remaining} attempts remaining"
            )
        new_timeout = self.healthcheck.timeout * pow(
            self.healthcheck.backoff_factor, self.attempts
        )
        self.healthcheck_deadline = now + new_timeout
        self.attempts += 1
        if self.attempts <= self.healthcheck.max_attempts:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "healthcheck: checking ..."
            )
            self.send(OscMessage(*self.healthcheck.request_pattern))
            return
        self._disconnect(panicked=True)

    def _server_factory(self, ip_address, port):
        server = self.Server(
            (self.ip_address, self.port), self.Handler, bind_and_activate=False
        )
        server.osc_protocol = self
        return server

    ### PUBLIC METHODS ###

    def activate_healthcheck(self) -> None:
        if self._activate_healthcheck():
            cast(HealthCheck, self.healthcheck).active = True

    def connect(
        self, ip_address: str, port: int, *, healthcheck: Optional[HealthCheck] = None
    ):
        if self.status != BootStatus.OFFLINE:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "already connected!"
            )
            raise OscProtocolAlreadyConnected
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
            "connecting ..."
        )
        self._setup(ip_address, port, healthcheck)
        self.healthcheck_deadline = time.time()
        self.boot_future = concurrent.futures.Future()
        self.exit_future = concurrent.futures.Future()
        self.osc_server = self._server_factory(ip_address, port)
        self.osc_server_thread = threading.Thread(target=self.osc_server.serve_forever)
        self.osc_server_thread.daemon = True
        self.osc_server_thread.start()
        if not self.healthcheck:
            self._on_connect(boot_future=self.boot_future)

    def disconnect(self) -> None:
        if self.status != BootStatus.ONLINE:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}/{self.name or hex(id(self))}] "
                "already disconnected!"
            )
            return
        self._disconnect()

    def register(
        self,
        pattern: Sequence[Union[str, float]],
        procedure: Callable[[OscMessage], Optional[Awaitable[None]]],
        *,
        failure_pattern: Optional[Sequence[Union[str, float]]] = None,
        once: bool = False,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict] = None,
    ) -> OscCallback:
        """
        Register a callback.
        """
        # Command queue prevents lock contention.
        self.command_queue.put(
            (
                "add",
                callback := self._register(
                    pattern,
                    procedure,
                    failure_pattern=failure_pattern,
                    once=once,
                    args=args,
                    kwargs=kwargs,
                ),
            )
        )
        return callback

    def send(self, message: Union[OscBundle, OscMessage, SequenceABC, str]) -> None:
        try:
            self.osc_server.socket.sendto(
                self._send(message),
                (self.ip_address, self.port),
            )
        except OSError:
            # print(message)
            raise

    def unregister(self, callback: OscCallback) -> None:
        """
        Unregister a callback.
        """
        # Command queue prevents lock contention.
        self.command_queue.put(("remove", callback))



import itertools
import random

import pytest
import uqbar.io

from supriya.utils._intervals import IntervalTreeDriverEx
from supriya.utils.intervals import Interval, IntervalTree, IntervalTreeDriver, Moment


def make_expected_start_offsets(range_=10, intervals=None):
    if not intervals:
        intervals = make_intervals()
    actual_offsets = set()
    for interval in intervals:
        actual_offsets.add(float(interval.start_offset))
    actual_offsets = sorted(actual_offsets)
    print("    O:", actual_offsets)
    offsets = {}
    for offset in range(-1, range_ + 1):
        before = [_ for _ in actual_offsets if _ < offset]
        if before:
            before = before[-1]
        else:
            before = None
        after = [_ for _ in actual_offsets if _ > offset]
        if after:
            after = after[0]
        else:
            after = None
        offsets[offset] = (before, after)
    return offsets


def make_moment_fixtures(range_=10, intervals=None):
    if not intervals:
        intervals = make_intervals()
    fixtures = {}
    for offset in range(range_):
        overlaps, starts, stops = [], [], []
        for interval in intervals:
            if interval.start_offset == offset:
                starts.append(interval)
            elif interval.stop_offset == offset:
                stops.append(interval)
            elif interval.start_offset < offset < interval.stop_offset:
                overlaps.append(interval)
        overlaps.sort()
        starts.sort()
        stops.sort()
        fixtures[offset] = (overlaps, starts, stops)
    return fixtures


def make_intervals():
    return [
        Interval(start_offset=0, stop_offset=3),
        Interval(start_offset=1, stop_offset=3),
        Interval(start_offset=1, stop_offset=2),
        Interval(start_offset=2, stop_offset=5),
        Interval(start_offset=6, stop_offset=9),
    ]


def make_interval_tree(accelerated, populated=True, intervals=None):
    if populated and not intervals:
        intervals = make_intervals()
    interval_tree = IntervalTree(intervals=intervals, accelerated=accelerated)
    if accelerated:
        assert isinstance(interval_tree._driver, IntervalTreeDriverEx)
    else:
        assert isinstance(interval_tree._driver, IntervalTreeDriver)
    return interval_tree


def make_random_intervals(count=10, range_=10):
    indices = list(range(range_))
    intervals = []
    for _ in range(count):
        random.shuffle(indices)
        start_offset, stop_offset = sorted(indices[:2])
        interval = Interval(start_offset=start_offset, stop_offset=stop_offset)
        intervals.append(interval)
    return intervals


def make_target_intervals(range_=10):
    indices = list(range(range_))
    intervals = []
    for pair in itertools.permutations(indices, 2):
        start_offset, stop_offset = sorted(pair)
        target_interval = Interval(start_offset=start_offset, stop_offset=stop_offset)
        intervals.append(target_interval)
    return intervals


@pytest.mark.parametrize("accelerated", [True, False])
def test___contains__(accelerated):
    intervals = make_intervals()
    interval_tree = make_interval_tree(accelerated=accelerated, populated=True)
    assert intervals[0] in interval_tree
    assert Interval(-1, 100) not in interval_tree
    interval_tree.remove(intervals[-1])
    assert intervals[-1] not in interval_tree


@pytest.mark.parametrize("accelerated", [True, False])
def test___getitem__(accelerated):
    interval_tree = make_interval_tree(accelerated=accelerated, populated=True)
    assert interval_tree[-1] == Interval(6, 9)
    assert [interval for interval in interval_tree[:3]] == [
        Interval(start_offset=0, stop_offset=3),
        Interval(start_offset=1, stop_offset=2),
        Interval(start_offset=1, stop_offset=3),
    ]


@pytest.mark.parametrize("accelerated", [True, False])
def test___init__(accelerated):
    make_interval_tree(accelerated=accelerated, populated=False)
    make_interval_tree(accelerated=accelerated, populated=True)


@pytest.mark.parametrize("accelerated", [True, False])
def test___iter__(accelerated):
    interval_tree = make_interval_tree(accelerated=accelerated, populated=True)
    assert [interval for interval in interval_tree] == [
        Interval(start_offset=0, stop_offset=3),
        Interval(start_offset=1, stop_offset=2),
        Interval(start_offset=1, stop_offset=3),
        Interval(start_offset=2, stop_offset=5),
        Interval(start_offset=6, stop_offset=9),
    ]
    iterator = iter(interval_tree)
    assert next(iterator) == Interval(start_offset=0, stop_offset=3)


@pytest.mark.parametrize("accelerated", [True, False])
def test___len__(accelerated):
    interval_tree = make_interval_tree(accelerated=accelerated, populated=False)
    assert len(interval_tree) == 0
    interval_tree = make_interval_tree(accelerated=accelerated, populated=True)
    assert len(interval_tree) == 5


@pytest.mark.parametrize("accelerated", [True, False])
def test___setitem__(accelerated):
    interval_tree = make_interval_tree(accelerated=accelerated, populated=True)
    interval_tree[-1] = Interval(-1, 4)
    assert [interval for interval in interval_tree] == [
        Interval(start_offset=-1, stop_offset=4),
        Interval(start_offset=0, stop_offset=3),
        Interval(start_offset=1, stop_offset=2),
        Interval(start_offset=1, stop_offset=3),
        Interval(start_offset=2, stop_offset=5),
    ]
    interval_tree[:3] = [Interval(100, 200)]
    assert [interval for interval in interval_tree] == [
        Interval(start_offset=1, stop_offset=3),
        Interval(start_offset=2, stop_offset=5),
        Interval(start_offset=100, stop_offset=200),
    ]


@pytest.mark.parametrize("accelerated", [True, False])
def test___sub__(accelerated):
    interval_tree = make_interval_tree(
        accelerated=accelerated,
        intervals=[Interval(0, 16), Interval(5, 12), Interval(-2, 8)],
    )
    interval = Interval(5, 10)
    result = interval_tree - interval
    assert result[:] == [
        Interval(-2, 5),
        Interval(0, 5),
        Interval(10, 12),
        Interval(10, 16),
    ]


@pytest.mark.parametrize("accelerated", [True, False])
def test_find_intersection_with_offset(accelerated):
    iterations = 10
    count, range_ = 10, 15
    for i in range(iterations):
        print("Iteration:", i)
        intervals = make_random_intervals(count=count, range_=range_)
        interval_tree = make_interval_tree(accelerated=accelerated, intervals=intervals)
        optimized = 0.0
        brute_force = 0.0
        for offset in range(range_):
            with uqbar.io.Timer() as timer:
                found_by_search = set(interval_tree.find_intersection(offset))
                optimized += timer.elapsed_time
            with uqbar.io.Timer() as timer:
                found_by_brute_force = set()
                for _ in interval_tree:
                    if _.start_offset <= offset < _.stop_offset:
                        found_by_brute_force.add(_)
                brute_force += timer.elapsed_time
            assert found_by_search == found_by_brute_force
        factor = "{:0.6f}".format(optimized / brute_force) if brute_force else "NaN"
        print(f"D: {factor} O: {optimized} B: {brute_force}")


@pytest.mark.parametrize("accelerated", [True, False])
def test_find_intersection_with_interval(accelerated):
    iterations = 10
    count, range_ = 10, 15
    target_intervals = make_target_intervals(range_=range_)
    for i in range(iterations):
        print("Iteration:", i)
        intervals = make_random_intervals(count=count, range_=range_)
        interval_tree = make_interval_tree(accelerated=accelerated, intervals=intervals)
        optimized = 0.0
        brute_force = 0.0
        for target_interval in target_intervals:
            with uqbar.io.Timer() as timer:
                found_by_search = set(interval_tree.find_intersection(target_interval))
                optimized += timer.elapsed_time
            with uqbar.io.Timer() as timer:
                found_by_brute_force = set()
                for _ in interval_tree:
                    if (
                        _.start_offset <= target_interval.start_offset
                        and target_interval.start_offset < _.stop_offset
                    ):
                        found_by_brute_force.add(_)
                    elif (
                        target_interval.start_offset <= _.start_offset
                        and _.start_offset < target_interval.stop_offset
                    ):
                        found_by_brute_force.add(_)
                brute_force += timer.elapsed_time
            assert found_by_search == found_by_brute_force
        factor = "{:0.6f}".format(optimized / brute_force) if brute_force else "NaN"
        print(f"D: {factor} O: {optimized} B: {brute_force}")


@pytest.mark.parametrize("accelerated", [True, False])
def test_find_intervals_starting_at(accelerated):
    iterations = 100
    count, range_ = 10, 15
    for i in range(iterations):
        print("Iteration:", i)
        intervals = make_random_intervals(count=count, range_=range_)
        interval_tree = make_interval_tree(accelerated=accelerated, intervals=intervals)
        for offset in range(range_):
            found_by_search = set(interval_tree.find_intervals_starting_at(offset))
            found_by_brute_force = set()
            for _ in interval_tree:
                if _.start_offset == offset:
                    found_by_brute_force.add(_)
            assert found_by_search == found_by_brute_force


@pytest.mark.parametrize("accelerated", [True, False])
def test_find_intervals_stopping_at(accelerated):
    iterations = 100
    count, range_ = 10, 15
    for i in range(iterations):
        print("Iteration:", i)
        intervals = make_random_intervals(count=count, range_=range_)
        interval_tree = make_interval_tree(accelerated=accelerated, intervals=intervals)
        for offset in range(range_):
            found_by_search = set(interval_tree.find_intervals_stopping_at(offset))
            found_by_brute_force = set()
            for _ in interval_tree:
                if _.stop_offset == offset:
                    found_by_brute_force.add(_)
            assert found_by_search == found_by_brute_force


@pytest.mark.parametrize("accelerated", [True, False])
def test_get_moment_at(accelerated):
    iterations = 100
    count, range_ = 10, 15
    for i in range(iterations):
        print("Iteration:", i)
        intervals = make_random_intervals(count=count, range_=range_)
        interval_tree = make_interval_tree(accelerated=accelerated, intervals=intervals)
        fixtures = make_moment_fixtures(range_=range_, intervals=intervals)
        for offset in range(range_):
            overlaps, starts, stops = fixtures[offset]
            expected = Moment(
                overlap_intervals=overlaps,
                start_offset=offset,
                start_intervals=starts,
                stop_intervals=stops,
                interval_tree=interval_tree,
            )
            actual = interval_tree.get_moment_at(offset)
            assert expected.interval_tree is actual.interval_tree
            assert expected.start_offset == actual.start_offset
            assert expected.start_intervals == actual.start_intervals
            assert expected.stop_intervals == actual.stop_intervals
            assert expected.overlap_intervals == actual.overlap_intervals


@pytest.mark.parametrize("accelerated", [True, False])
def test_get_start_offset(accelerated):
    iterations = 100
    count, range_ = 10, 15
    for i in range(iterations):
        print("Iteration:", i)
        intervals = make_random_intervals(count=count, range_=range_)
        interval_tree = make_interval_tree(accelerated=accelerated, intervals=intervals)
        expected_offsets = make_expected_start_offsets(
            range_=range_, intervals=intervals
        )
        for interval in sorted(intervals):
            print("    Interval:", interval)
        for offset in range(-1, range_ + 1):
            print("    Offset:", offset)
            print("        :", expected_offsets[offset])
            expected_before, expected_after = expected_offsets[offset]
            actual_before = interval_tree.get_start_offset_before(offset)
            actual_after = interval_tree.get_start_offset_after(offset)
            assert expected_before == actual_before, offset
            assert expected_after == actual_after, offset


@pytest.mark.parametrize("accelerated", [True, False])
def test_index(accelerated):
    intervals = make_intervals()
    interval_tree = make_interval_tree(accelerated=accelerated, populated=True)
    assert interval_tree.index(intervals[0]) == 0
    assert interval_tree.index(intervals[1]) == 2
    assert interval_tree.index(intervals[2]) == 1
    assert interval_tree.index(intervals[3]) == 3
    assert interval_tree.index(intervals[4]) == 4
    with pytest.raises(ValueError):
        interval = Interval(-100, 100)
        interval_tree.index(interval)


@pytest.mark.parametrize("accelerated", [True, False])
def test_insert(accelerated):
    interval_tree = make_interval_tree(accelerated=accelerated, populated=False)
    interval_tree.add(Interval(1, 3))
    interval_tree.update((Interval(0, 4), Interval(2, 6)))
    assert interval_tree[:] == [
        Interval(start_offset=0, stop_offset=4),
        Interval(start_offset=1, stop_offset=3),
        Interval(start_offset=2, stop_offset=6),
    ]


@pytest.mark.parametrize("accelerated", [True, False])
def test_iterate_moments(accelerated):
    interval_tree = make_interval_tree(accelerated=accelerated, populated=True)
    moments = list(interval_tree.iterate_moments())
    assert [x.start_offset for x in moments] == [0, 1, 2, 6]
    moments = list(interval_tree.iterate_moments(reverse=True))
    assert [x.start_offset for x in moments] == [6, 2, 1, 0]


@pytest.mark.parametrize("accelerated", [True, False])
def test_remove(accelerated):
    intervals = make_intervals()
    interval_tree = make_interval_tree(accelerated=accelerated, populated=True)
    assert list(interval_tree) == sorted(intervals)
    with pytest.raises(ValueError):
        interval_tree.remove(intervals[1:-1])
    assert list(interval_tree) == sorted(intervals)
    for interval in interval_tree[1:-1]:
        interval_tree.remove(interval)
    assert interval_tree[:] == [
        Interval(start_offset=0, stop_offset=3),
        Interval(start_offset=6, stop_offset=9),
    ]


def test_get_offset_after():
    intervals = [
        Interval(0, 3),
        Interval(1, 3),
        Interval(1, 2),
        Interval(2, 5),
        Interval(5, 10),
        Interval(5, 12),
        Interval(6, 9),
        Interval(13, 15),
    ]
    expected = [
        (-2, 0.0),
        (-1, 0.0),
        (0, 1.0),
        (1, 2.0),
        (2, 3.0),
        (3, 5.0),
        (4, 5.0),
        (5, 6.0),
        (6, 9.0),
        (7, 9.0),
        (8, 9.0),
        (9, 10.0),
        (10, 12.0),
        (11, 12.0),
        (12, 13.0),
        (13, 15.0),
        (14, 15.0),
        (15, None),
        (16, None),
    ]
    for _ in range(10):
        interval_tree = IntervalTree(intervals)
        actual = [(i, interval_tree.get_offset_after(i)) for i in range(-2, 17)]
        assert actual == expected
        random.shuffle(intervals)



import queue
import sys
from typing import Dict, List, Optional

from .ephemera import Event

if sys.version_info >= (3, 9):
    _EventQueueBase = queue.PriorityQueue[Event]
else:
    _EventQueueBase = queue.PriorityQueue


class EventQueue(_EventQueueBase):
    ### PRIVATE METHODS ###

    def _init(self, maxsize: Optional[int]) -> None:
        self.queue: List[Event] = []
        self.flags: Dict[Event, bool] = {}

    def _put(self, event: Event) -> None:
        self.flags[event] = True
        super()._put(event)

    def _get(self) -> Event:
        while self.queue:
            if not self.flags.pop((event := super()._get()), None):
                continue
            return event
        raise queue.Empty

    ### PUBLIC METHODS ###

    def clear(self) -> None:
        with self.mutex:
            self._init(None)

    def peek(self) -> Event:
        with self.mutex:
            self._put(event := self._get())
        return event

    def remove(self, event: Event) -> None:
        with self.mutex:
            self.flags.pop(event, None)



import datetime
import os

import supriya

### SPHINX ###

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    # "sphinx.ext.viewcode",
    "sphinxext.opengraph",
    "uqbar.sphinx.api",
    "uqbar.sphinx.book",
    "uqbar.sphinx.inheritance",
    "sphinx_immaterial",
]

add_module_names = False
copyright = f"2014-{datetime.date.today().year}, Joséphine Wolf Oberholtzer"
exclude_patterns = []
htmlhelp_basename = "Supriyadoc"
language = "en"
master_doc = "index"
project = "Supriya"
pygments_style = "sphinx"
source_suffix = ".rst"
templates_path = ["_templates"]
version = release = supriya.__version__

### GRAPHVIZ ###

graphviz_dot_args = ["-s32"]
graphviz_output_format = "svg"

### INTERSPHINX ###

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "uqbar": ("http://supriya-project.github.io/uqbar", None),
}

### OPENGRAPH ###

ogp_site_url = "https://supriya-project.github.io/supriya/"

### TODO ###

todo_include_todos = True

### UQBAR API ###

uqbar_api_member_documenter_classes = [
    "uqbar.apis.FunctionDocumenter",
    "uqbar.apis.ImmaterialClassDocumenter",
]
uqbar_api_module_documenter_class = "uqbar.apis.ImmaterialModuleDocumenter"
uqbar_api_omit_root = True
uqbar_api_root_documenter_class = "uqbar.apis.SummarizingRootDocumenter"
uqbar_api_source_paths = supriya.__path__
uqbar_api_title = "Supriya API Reference"

### UQBAR BOOK ###

uqbar_book_console_setup = ["import supriya"]
uqbar_book_console_teardown = """\
import asyncio

async def shutdown():
    for server in tuple(supriya.Server._contexts):
        if asyncio.iscoroutine(result := server._shutdown()):
            await result

asyncio.run(shutdown())
""".splitlines()
uqbar_book_extensions = [
    "uqbar.book.extensions.GraphExtension",
    "supriya.ext.book.PlayExtension",
    "supriya.ext.book.PlotExtension",
]
uqbar_book_strict = os.environ.get("CI") == "true"
uqbar_book_use_black = True
uqbar_book_use_cache = False

### THEME ###

html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.4/css/all.min.css"
]
html_domain_indices = True
html_favicon = "favicon.ico"
html_js_files = []
html_logo = "icon.svg"
html_static_path = ["_static"]
html_theme = "sphinx_immaterial"
html_theme_options = {
    "icon": {"repo": "fontawesome/brands/github"},
    "site_url": "https://supriya-project.github.io/supriya/",
    "repo_url": "https://github.com/supriya-project/supriya/",
    "repo_name": "supriya",
    "repo_type": "github",
    "edit_uri": "blob/main/docs",
    "globaltoc_collapse": False,
    "features": [
        # "header.autohide",
        # "navigation.expand",
        # "navigation.instant",
        # "navigation.sections",
        "navigation.tabs",
        "navigation.top",
        # "search.highlight",
        # "search.share",
        # "toc.integrate",
    ],
    "palette": [
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "blue-grey",
            "accent": "lime",
            "toggle": {
                "icon": "material/toggle-switch",
                "name": "Switch to light mode",
            },
        },
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "indigo",
            "accent": "teal",
            "toggle": {
                "icon": "material/toggle-switch-off-outline",
                "name": "Switch to dark mode",
            },
        },
    ],
    "version_dropdown": False,
}
html_title = "Supriya"
html_use_index = True
object_description_options = [
    ("py:.*", dict(include_fields_in_toc=False)),  # Hide "Parameters" in TOC
    ("py:parameter", dict(include_in_toc=False)),  # Hide "p" parameter entries in TOC
    ("py:exception", {"toc_icon_class": "data", "toc_icon_text": "X"}),
]



class AllocationError(Exception):
    pass


class ContextError(Exception):
    pass


class InvalidCalculationRate(ContextError):
    pass


class InvalidMoment(ContextError):
    pass


class MomentClosed(ContextError):
    pass



import pytest

from supriya.patterns import ChainPattern, EventPattern, NoteEvent, SequencePattern
from supriya.patterns.testutils import MockUUID as M
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, input_a, input_b1, input_b2, input_c, expected, is_infinite",
    [
        (
            None,
            SequencePattern([1, 2, 3]),
            SequencePattern([4, 5]),
            SequencePattern([7, 8, 9]),
            SequencePattern([10, 11]),
            [NoteEvent(M("A"), a=1, b=7, c=10), NoteEvent(M("B"), a=2, b=8, c=11)],
            False,
        ),
        (
            None,
            SequencePattern([1, 2, 3], None),
            SequencePattern([4, 5], None),
            SequencePattern([7, 8, 9]),
            SequencePattern([10, 11]),
            [NoteEvent(M("A"), a=1, b=7, c=10), NoteEvent(M("B"), a=2, b=8, c=11)],
            False,
        ),
        (
            None,
            SequencePattern([1, 2, 3], None),
            SequencePattern([4, 5], None),
            SequencePattern([7, 8, 9], None),
            SequencePattern([10, 11], None),
            [
                NoteEvent(M("A"), a=1, b=7, c=10),
                NoteEvent(M("B"), a=2, b=8, c=11),
                NoteEvent(M("C"), a=3, b=9, c=10),
                NoteEvent(M("D"), a=1, b=7, c=11),
                NoteEvent(M("E"), a=2, b=8, c=10),
                NoteEvent(M("F"), a=3, b=9, c=11),
            ],
            True,
        ),
    ],
)
def test(stop_at, input_a, input_b1, input_b2, input_c, expected, is_infinite):
    pattern = ChainPattern(
        EventPattern(a=input_a, b=input_b1),
        EventPattern(b=input_b2),
        EventPattern(c=input_c),
    )
    run_pattern_test(pattern, expected, is_infinite, stop_at)



import pytest

from supriya.patterns import (
    CompositeEvent,
    EventPattern,
    GroupAllocateEvent,
    GroupPattern,
    NodeFreeEvent,
    NoteEvent,
    NullEvent,
    ParallelPattern,
    SequencePattern,
)
from supriya.patterns.testutils import MockUUID as M
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, patterns, expected, is_infinite",
    [
        (
            None,
            [
                EventPattern(frequency=SequencePattern([440, 550, 660])),
                EventPattern(frequency=SequencePattern([777, 888, 999])),
            ],
            [
                CompositeEvent(
                    [
                        NoteEvent(M("A"), delta=0.0, frequency=440),
                        NoteEvent(M("B"), delta=0.0, frequency=777),
                    ],
                    delta=1.0,
                ),
                CompositeEvent(
                    [
                        NoteEvent(M("C"), delta=0.0, frequency=550),
                        NoteEvent(M("D"), delta=0.0, frequency=888),
                    ],
                    delta=1.0,
                ),
                CompositeEvent(
                    [
                        NoteEvent(M("E"), delta=0.0, frequency=660),
                        NoteEvent(M("F"), delta=0.0, frequency=999),
                    ],
                    delta=1.0,
                ),
            ],
            False,
        ),
        (
            None,
            [
                EventPattern(x=SequencePattern([1, 2, 3]), delta=1.0),
                EventPattern(y=SequencePattern([1, 2]), delta=1.5),
            ],
            [
                CompositeEvent(
                    [
                        NoteEvent(M("A"), delta=0.0, x=1),
                        NoteEvent(M("B"), delta=0.0, y=1),
                    ],
                    delta=1.0,
                ),
                NoteEvent(M("C"), delta=0.5, x=2),
                NoteEvent(M("D"), delta=0.5, y=2),
                NoteEvent(M("E"), delta=1.0, x=3),
            ],
            False,
        ),
        (
            1,
            [
                EventPattern(x=SequencePattern([1, 2, 3]), delta=1.0),
                EventPattern(y=SequencePattern([1, 2]), delta=1.5),
            ],
            [
                CompositeEvent(
                    [
                        NoteEvent(M("A"), delta=0.0, x=1),
                        NoteEvent(M("B"), delta=0.0, y=1),
                    ],
                    delta=1.0,
                )
            ],
            False,
        ),
        (
            None,
            [
                GroupPattern(EventPattern(x=SequencePattern([1, 2, 3]), delta=1.0)),
                GroupPattern(EventPattern(y=SequencePattern([1, 2]), delta=1.5)),
            ],
            [
                CompositeEvent(
                    [
                        CompositeEvent([GroupAllocateEvent(M("A"))]),
                        NoteEvent(M("B"), delta=0.0, target_node=M("A"), x=1),
                        CompositeEvent([GroupAllocateEvent(M("C"))]),
                        NoteEvent(M("D"), delta=0.0, target_node=M("C"), y=1),
                    ],
                    delta=1.0,
                ),
                NoteEvent(M("E"), delta=0.5, target_node=M("A"), x=2),
                NoteEvent(M("F"), delta=0.5, target_node=M("C"), y=2),
                NoteEvent(M("G"), delta=1.0, target_node=M("A"), x=3),
                CompositeEvent(
                    [
                        CompositeEvent([NullEvent(delta=0.25), NodeFreeEvent(M("A"))]),
                        CompositeEvent([NullEvent(delta=0.25), NodeFreeEvent(M("C"))]),
                    ]
                ),
            ],
            False,
        ),
        (
            1,
            [
                GroupPattern(EventPattern(x=SequencePattern([1, 2, 3]), delta=1.0)),
                GroupPattern(EventPattern(y=SequencePattern([1, 2]), delta=1.5)),
            ],
            [
                CompositeEvent(
                    [
                        CompositeEvent([GroupAllocateEvent(M("A"))]),
                        NoteEvent(M("B"), delta=0.0, target_node=M("A"), x=1),
                        CompositeEvent([GroupAllocateEvent(M("C"))]),
                        NoteEvent(M("D"), delta=0.0, target_node=M("C"), y=1),
                    ],
                    delta=1.0,
                ),
                CompositeEvent(
                    [NullEvent(delta=0.25), NodeFreeEvent(M("A"))], delta=0.5
                ),
                CompositeEvent([NullEvent(delta=0.25), NodeFreeEvent(M("C"))]),
            ],
            False,
        ),
    ],
)
def test(stop_at, patterns, expected, is_infinite):
    pattern = ParallelPattern(patterns)
    run_pattern_test(pattern, expected, is_infinite, stop_at)



import collections
import datetime
import enum
import struct
import time
from collections.abc import Sequence as SequenceABC
from typing import List

from uqbar.objects import get_repr

from ..utils import group_by_count

BUNDLE_PREFIX = b"#bundle\x00"
IMMEDIATELY = struct.pack(">Q", 1)
NTP_TIMESTAMP_TO_SECONDS = 1.0 / 2.0**32.0
SECONDS_TO_NTP_TIMESTAMP = 2.0**32.0
SYSTEM_EPOCH = datetime.date(*time.gmtime(0)[0:3])
NTP_EPOCH = datetime.date(1900, 1, 1)
NTP_DELTA = (SYSTEM_EPOCH - NTP_EPOCH).days * 24 * 3600


def format_datagram(datagram: bytes) -> str:
    result: List[str] = ["size {}".format(len(datagram))]
    index = 0
    while index < len(datagram):
        hex_blocks = []
        ascii_block = ""
        for chunk in group_by_count(datagram[index : index + 16], 4):
            hex_block = []
            for byte in chunk:
                if 31 < int(byte) < 127:
                    char = chr(int(byte))
                else:
                    char = "."
                ascii_block += char
                hexed = hex(byte)[2:].zfill(2)
                hex_block.append(hexed)
            hex_blocks.append(" ".join(hex_block))
        line = "{: >4}   ".format(index)
        line += "{: <53}".format("  ".join(hex_blocks))
        line += "|{}|".format(ascii_block)
        result.append(line)
        index += 16
    return "\n".join(result)


class OscMessage:
    """
    An OSC message.

    .. container:: example

        ::

            >>> from supriya.osc import OscMessage
            >>> osc_message = OscMessage("/g_new", 0, 0)
            >>> osc_message
            OscMessage('/g_new', 0, 0)

        ::

            >>> datagram = osc_message.to_datagram()
            >>> OscMessage.from_datagram(datagram)
            OscMessage('/g_new', 0, 0)

        ::

            >>> print(osc_message)
            size 20
               0   2f 67 5f 6e  65 77 00 00  2c 69 69 00  00 00 00 00   |/g_new..,ii.....|
              16   00 00 00 00                                          |....|

    .. container:: example

        ::

            >>> osc_message = OscMessage("/foo", True, [None, [3.25]], OscMessage("/bar"))
            >>> osc_message
            OscMessage('/foo', True, [None, [3.25]], OscMessage('/bar'))

        ::

            >>> datagram = osc_message.to_datagram()
            >>> OscMessage.from_datagram(datagram)
            OscMessage('/foo', True, [None, [3.25]], OscMessage('/bar'))

        ::

            >>> print(osc_message)
            size 40
               0   2f 66 6f 6f  00 00 00 00  2c 54 5b 4e  5b 66 5d 5d   |/foo....,T[N[f]]|
              16   62 00 00 00  40 50 00 00  00 00 00 0c  2f 62 61 72   |b...@P....../bar|
              32   00 00 00 00  2c 00 00 00                             |....,...|

    .. container:: example

        ::

            >>> osc_message = supriya.osc.OscMessage(
            ...     "/foo",
            ...     1,
            ...     2.5,
            ...     supriya.osc.OscBundle(
            ...         contents=(
            ...             supriya.osc.OscMessage("/bar", "baz", 3.0),
            ...             supriya.osc.OscMessage("/ffff", False, True, None),
            ...         )
            ...     ),
            ...     ["a", "b", ["c", "d"]],
            ... )
            >>> osc_message
            OscMessage('/foo', 1, 2.5, OscBundle(
                contents=(
                    OscMessage('/bar', 'baz', 3.0),
                    OscMessage('/ffff', False, True, None),
                ),
            ), ['a', 'b', ['c', 'd']])

        ::

            >>> datagram = osc_message.to_datagram()
            >>> OscMessage.from_datagram(datagram)
            OscMessage('/foo', 1, 2.5, OscBundle(
                contents=(
                    OscMessage('/bar', 'baz', 3.0),
                    OscMessage('/ffff', False, True, None),
                ),
            ), ['a', 'b', ['c', 'd']])

        ::

            >>> print(osc_message)
            size 112
               0   2f 66 6f 6f  00 00 00 00  2c 69 66 62  5b 73 73 5b   |/foo....,ifb[ss[|
              16   73 73 5d 5d  00 00 00 00  00 00 00 01  40 20 00 00   |ss]]........@ ..|
              32   00 00 00 3c  23 62 75 6e  64 6c 65 00  00 00 00 00   |...<#bundle.....|
              48   00 00 00 01  00 00 00 14  2f 62 61 72  00 00 00 00   |......../bar....|
              64   2c 73 66 00  62 61 7a 00  40 40 00 00  00 00 00 10   |,sf.baz.@@......|
              80   2f 66 66 66  66 00 00 00  2c 46 54 4e  00 00 00 00   |/ffff...,FTN....|
              96   61 00 00 00  62 00 00 00  63 00 00 00  64 00 00 00   |a...b...c...d...|
    """

    ### INITIALIZER ###

    def __init__(self, address, *contents) -> None:
        if isinstance(address, enum.Enum):
            address = address.value
        if not isinstance(address, (str, int)):
            raise ValueError(f"address must be int or str, got {address}")
        self.address = address
        self.contents = tuple(contents)

    ### SPECIAL METHODS ###

    def __eq__(self, other) -> bool:
        if type(self) is not type(other):
            return False
        if self.address != other.address:
            return False
        if self.contents != other.contents:
            return False
        return True

    def __repr__(self) -> str:
        return "{}({})".format(
            type(self).__name__,
            ", ".join(repr(_) for _ in [self.address, *self.contents]),
        )

    def __str__(self) -> str:
        return format_datagram(bytearray(self.to_datagram()))

    ### PRIVATE METHODS ###

    @staticmethod
    def _decode_blob(data):
        actual_length, remainder = struct.unpack(">I", data[:4])[0], data[4:]
        padded_length = actual_length
        if actual_length % 4 != 0:
            padded_length = (actual_length // 4 + 1) * 4
        return remainder[:padded_length][:actual_length], remainder[padded_length:]

    @staticmethod
    def _decode_string(data):
        actual_length = data.index(b"\x00")
        padded_length = (actual_length // 4 + 1) * 4
        return str(data[:actual_length], "ascii"), data[padded_length:]

    @staticmethod
    def _encode_string(value):
        result = bytes(value + "\x00", "ascii")
        if len(result) % 4 != 0:
            width = (len(result) // 4 + 1) * 4
            result = result.ljust(width, b"\x00")
        return result

    @staticmethod
    def _encode_blob(value):
        result = bytes(struct.pack(">I", len(value)) + value)
        if len(result) % 4 != 0:
            width = (len(result) // 4 + 1) * 4
            result = result.ljust(width, b"\x00")
        return result

    @classmethod
    def _encode_value(cls, value):
        if hasattr(value, "to_datagram"):
            value = bytearray(value.to_datagram())
        elif isinstance(value, enum.Enum):
            value = value.value
        type_tags, encoded_value = "", b""
        if isinstance(value, (bytearray, bytes)):
            type_tags += "b"
            encoded_value = cls._encode_blob(value)
        elif isinstance(value, str):
            type_tags += "s"
            encoded_value = cls._encode_string(value)
        elif isinstance(value, bool):
            type_tags += "T" if value else "F"
        elif isinstance(value, float):
            type_tags += "f"
            encoded_value += struct.pack(">f", value)
        elif isinstance(value, int):
            type_tags += "i"
            encoded_value += struct.pack(">i", value)
        elif value is None:
            type_tags += "N"
        elif isinstance(value, SequenceABC):
            type_tags += "["
            for sub_value in value:
                sub_type_tags, sub_encoded_value = cls._encode_value(sub_value)
                type_tags += sub_type_tags
                encoded_value += sub_encoded_value
            type_tags += "]"
        else:
            message = "Cannot encode {!r}".format(value)
            raise TypeError(message)
        return type_tags, encoded_value

    ### PUBLIC METHODS ###

    def to_datagram(self) -> bytes:
        # address can be a string or (in SuperCollider) an int
        if isinstance(self.address, str):
            encoded_address = self._encode_string(self.address)
        else:
            encoded_address = struct.pack(">i", self.address)
        encoded_type_tags = ","
        encoded_contents = b""
        for value in self.contents or ():
            type_tags, encoded_value = self._encode_value(value)
            encoded_type_tags += type_tags
            encoded_contents += encoded_value
        return (
            encoded_address + self._encode_string(encoded_type_tags) + encoded_contents
        )

    @classmethod
    def from_datagram(cls, datagram):
        remainder = datagram
        address, remainder = cls._decode_string(remainder)
        type_tags, remainder = cls._decode_string(remainder)
        contents = []
        array_stack = [contents]
        for type_tag in type_tags[1:]:
            if type_tag == "i":
                value, remainder = struct.unpack(">i", remainder[:4])[0], remainder[4:]
                array_stack[-1].append(value)
            elif type_tag == "f":
                value, remainder = struct.unpack(">f", remainder[:4])[0], remainder[4:]
                array_stack[-1].append(value)
            elif type_tag == "d":
                value, remainder = struct.unpack(">d", remainder[:8])[0], remainder[8:]
                array_stack[-1].append(value)
            elif type_tag == "s":
                value, remainder = cls._decode_string(remainder)
                array_stack[-1].append(value)
            elif type_tag == "b":
                value, remainder = cls._decode_blob(remainder)
                for class_ in (OscBundle, OscMessage):
                    try:
                        value = class_.from_datagram(value)
                        break
                    except Exception:
                        pass
                array_stack[-1].append(value)
            elif type_tag == "T":
                array_stack[-1].append(True)
            elif type_tag == "F":
                array_stack[-1].append(False)
            elif type_tag == "N":
                array_stack[-1].append(None)
            elif type_tag == "[":
                array = []
                array_stack[-1].append(array)
                array_stack.append(array)
            elif type_tag == "]":
                array_stack.pop()
            else:
                raise RuntimeError(f"Unable to parse type {type_tag!r}")
        return cls(address, *contents)

    def to_list(self):
        result = [self.address]
        for x in self.contents:
            if hasattr(x, "to_list"):
                result.append(x.to_list())
            else:
                result.append(x)
        return result


class OscBundle:
    """
    An OSC bundle.

    ::

        >>> import supriya.osc
        >>> message_one = supriya.osc.OscMessage("/one", 1)
        >>> message_two = supriya.osc.OscMessage("/two", 2)
        >>> message_three = supriya.osc.OscMessage("/three", 3)

    ::

        >>> inner_bundle = supriya.osc.OscBundle(
        ...     timestamp=1401557034.5,
        ...     contents=(message_one, message_two),
        ... )
        >>> inner_bundle
        OscBundle(
            contents=(
                OscMessage('/one', 1),
                OscMessage('/two', 2),
            ),
            timestamp=1401557034.5,
        )

    ::

        >>> print(inner_bundle)
        size 56
           0   23 62 75 6e  64 6c 65 00  d7 34 8e aa  80 00 00 00   |#bundle..4......|
          16   00 00 00 10  2f 6f 6e 65  00 00 00 00  2c 69 00 00   |..../one....,i..|
          32   00 00 00 01  00 00 00 10  2f 74 77 6f  00 00 00 00   |......../two....|
          48   2c 69 00 00  00 00 00 02                             |,i......|

    ::

        >>> outer_bundle = supriya.osc.OscBundle(
        ...     contents=(inner_bundle, message_three),
        ... )
        >>> outer_bundle
        OscBundle(
            contents=(
                OscBundle(
                    contents=(
                        OscMessage('/one', 1),
                        OscMessage('/two', 2),
                    ),
                    timestamp=1401557034.5,
                ),
                OscMessage('/three', 3),
            ),
        )

    ::

        >>> print(outer_bundle)
        size 96
           0   23 62 75 6e  64 6c 65 00  00 00 00 00  00 00 00 01   |#bundle.........|
          16   00 00 00 38  23 62 75 6e  64 6c 65 00  d7 34 8e aa   |...8#bundle..4..|
          32   80 00 00 00  00 00 00 10  2f 6f 6e 65  00 00 00 00   |......../one....|
          48   2c 69 00 00  00 00 00 01  00 00 00 10  2f 74 77 6f   |,i........../two|
          64   00 00 00 00  2c 69 00 00  00 00 00 02  00 00 00 10   |....,i..........|
          80   2f 74 68 72  65 65 00 00  2c 69 00 00  00 00 00 03   |/three..,i......|

    ::

        >>> datagram = outer_bundle.to_datagram()

    ::

        >>> decoded_bundle = supriya.osc.OscBundle.from_datagram(datagram)
        >>> decoded_bundle
        OscBundle(
            contents=(
                OscBundle(
                    contents=(
                        OscMessage('/one', 1),
                        OscMessage('/two', 2),
                    ),
                    timestamp=1401557034.5,
                ),
                OscMessage('/three', 3),
            ),
        )

    ::

        >>> decoded_bundle == outer_bundle
        True
    """

    ### INITIALIZER ###

    def __init__(self, timestamp=None, contents=None) -> None:
        prototype = (OscMessage, type(self))
        self.timestamp = timestamp
        contents = contents or ()
        for x in contents or ():
            if not isinstance(x, prototype):
                raise ValueError(contents)
        self.contents = tuple(contents)

    ### SPECIAL METHODS ###

    def __eq__(self, other) -> bool:
        if type(self) is not type(other):
            return False
        if self.timestamp != other.timestamp:
            return False
        if self.contents != other.contents:
            return False
        return True

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return format_datagram(bytearray(self.to_datagram()))

    ### PRIVATE METHODS ###

    @staticmethod
    def _decode_date(data):
        data, remainder = data[:8], data[8:]
        if data == IMMEDIATELY:
            return None, remainder
        date = (struct.unpack(">Q", data)[0] / SECONDS_TO_NTP_TIMESTAMP) - NTP_DELTA
        return date, remainder

    @staticmethod
    def _encode_date(seconds, realtime=True):
        if seconds is None:
            return IMMEDIATELY
        if realtime:
            seconds = seconds + NTP_DELTA
        if seconds >= 4294967296:  # 2**32
            seconds = seconds % 4294967296
        return struct.pack(">Q", int(seconds * SECONDS_TO_NTP_TIMESTAMP))

    ### PUBLIC METHODS ###

    @classmethod
    def from_datagram(cls, datagram):
        if not datagram.startswith(BUNDLE_PREFIX):
            raise ValueError("datagram is not a bundle")
        remainder = datagram[8:]
        timestamp, remainder = cls._decode_date(remainder)
        contents = []
        while len(remainder):
            length, remainder = struct.unpack(">i", remainder[:4])[0], remainder[4:]
            if remainder.startswith(BUNDLE_PREFIX):
                item = cls.from_datagram(remainder[:length])
            else:
                item = OscMessage.from_datagram(remainder[:length])
            contents.append(item)
            remainder = remainder[length:]
        osc_bundle = cls(timestamp=timestamp, contents=tuple(contents))
        return osc_bundle

    @classmethod
    def partition(cls, messages, timestamp=None):
        bundles = []
        contents = []
        message = collections.deque(messages)
        remaining = maximum = 8192 - len(BUNDLE_PREFIX) - 4
        while messages:
            message = messages.popleft()
            datagram = message.to_datagram()
            remaining -= len(datagram) + 4
            if remaining > 0:
                contents.append(message)
            else:
                bundles.append(cls(timestamp=timestamp, contents=contents))
                contents = [message]
                remaining = maximum
        if contents:
            bundles.append(cls(timestamp=timestamp, contents=contents))
        return bundles

    def to_datagram(self, realtime=True) -> bytes:
        datagram = BUNDLE_PREFIX
        datagram += self._encode_date(self.timestamp, realtime=realtime)
        for content in self.contents:
            content_datagram = content.to_datagram()
            datagram += struct.pack(">i", len(content_datagram))
            datagram += content_datagram
        return datagram

    def to_list(self):
        result = [self.timestamp]
        result.append([x.to_list() for x in self.contents])
        return result



from .core import UGen, param, ugen


@ugen(ar=True)
class FreqShift(UGen):
    """
    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> freq_shift = supriya.ugens.FreqShift.ar(
        ...     frequency=0,
        ...     phase=0,
        ...     source=source,
        ... )
        >>> freq_shift
        <FreqShift.ar()[0]>
    """

    source = param()
    frequency = param(0.0)
    phase = param(0.0)


@ugen(ar=True, channel_count=2, fixed_channel_count=True)
class Hilbert(UGen):
    """
    Applies the Hilbert transform.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> hilbert = supriya.ugens.Hilbert.ar(
        ...     source=source,
        ... )
        >>> hilbert
        <Hilbert.ar()>
    """

    source = param()


@ugen(ar=True)
class HilbertFIR(UGen):
    """
    Applies the Hilbert transform.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> hilbert_fir = supriya.ugens.HilbertFIR.ar(
        ...     buffer_id=23,
        ...     source=source,
        ... )
        >>> hilbert_fir
        <HilbertFIR.ar()[0]>
    """

    source = param()
    buffer_id = param()



import operator

import pytest

from supriya.patterns import BinaryOpPattern, SequencePattern, UnaryOpPattern


@pytest.mark.parametrize(
    "op, expr_one, expr_two, expected",
    [
        (
            operator.add,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.add, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.add,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.add, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.and_,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.and_, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.and_,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.and_, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.floordiv,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.floordiv, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.floordiv,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.floordiv, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.ge,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.le, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.ge,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.ge, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.gt,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.lt, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.gt,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.gt, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.le,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.ge, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.le,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.le, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.lshift,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.lshift, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.lshift,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.lshift, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.lt,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.gt, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.lt,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.lt, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.mod,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.mod, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.mod,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.mod, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.mul,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.mul, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.mul,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.mul, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.or_,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.or_, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.or_,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.or_, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.pow,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.pow, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.pow,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.pow, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.rshift,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.rshift, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.rshift,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.rshift, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.sub,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.sub, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.sub,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.sub, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.truediv,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.truediv, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.truediv,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.truediv, SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.xor,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern(operator.xor, 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.xor,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern(operator.xor, SequencePattern([1, 2, 3]), 23),
        ),
    ],
)
def test_binary_ops(op, expr_one, expr_two, expected):
    if isinstance(expected, Exception):
        with pytest.raises(expected):
            op(expr_one, expr_two)
    else:
        assert op(expr_one, expr_two) == expected


@pytest.mark.parametrize(
    "op, expr, expected",
    [
        (
            operator.abs,
            SequencePattern([1, 2, 3]),
            UnaryOpPattern[int](operator.abs, SequencePattern([1, 2, 3])),
        ),
        (
            operator.inv,
            SequencePattern([1, 2, 3]),
            UnaryOpPattern[int](operator.invert, SequencePattern([1, 2, 3])),
        ),
        (
            operator.neg,
            SequencePattern([1, 2, 3]),
            UnaryOpPattern[int](operator.neg, SequencePattern([1, 2, 3])),
        ),
        (
            operator.pos,
            SequencePattern([1, 2, 3]),
            UnaryOpPattern[int](operator.pos, SequencePattern([1, 2, 3])),
        ),
    ],
)
def test_unary_ops(op, expr, expected):
    if isinstance(expected, Exception):
        with pytest.raises(expected):
            op(expr)
    else:
        assert op(expr) == expected



from typing import List, Optional, Sequence, Tuple, Union

from uqbar.objects import get_repr

from .. import utils
from ..enums import DoneAction, EnvelopeShape
from .core import Parameter, UGen, UGenOperable, UGenVector, param, ugen


class Envelope:
    """
    An envelope.

    ::

        >>> from supriya.ugens import Envelope
        >>> envelope = Envelope()

    ::

        >>> list(envelope.serialize())
        [<0.0>, <2.0>, <-99.0>, <-99.0>, <1.0>, <1.0>, <1.0>, <0.0>, <0.0>, <1.0>, <1.0>, <0.0>]
    """

    def __init__(
        self,
        amplitudes: Sequence[Union[UGenOperable, float]] = (0, 1, 0),
        durations: Sequence[Union[UGenOperable, float]] = (1, 1),
        curves: Sequence[Union[EnvelopeShape, UGenOperable, float, str, None]] = (
            EnvelopeShape.LINEAR,
            EnvelopeShape.LINEAR,
        ),
        release_node: Optional[int] = None,
        loop_node: Optional[int] = None,
        offset: Union[UGenOperable, float] = 0.0,
    ) -> None:
        assert len(amplitudes) > 1
        assert len(durations) == (len(amplitudes) - 1)
        if isinstance(curves, (int, float, str, EnvelopeShape, UGenOperable)):
            curves = [curves]
        elif curves is None:
            curves = []
        self._release_node = release_node
        self._loop_node = loop_node
        self._offset = offset
        self._initial_amplitude = amplitudes[0]
        self._amplitudes: Tuple[Union[UGenOperable, float], ...] = tuple(amplitudes[1:])
        self._durations: Tuple[Union[UGenOperable, float], ...] = tuple(durations)
        curves_: List[Union[EnvelopeShape, UGenOperable, float]] = []
        for x in curves:
            if isinstance(x, (EnvelopeShape, UGenOperable)):
                curves_.append(x)
            elif isinstance(x, str) or x is None:
                curves_.append(EnvelopeShape.from_expr(x))
            else:
                curves_.append(float(x))
        self._curves: Tuple[Union[EnvelopeShape, UGenOperable, float], ...] = tuple(
            curves_
        )
        self._envelope_segments = tuple(
            utils.zip_cycled(self._amplitudes, self._durations, self._curves)
        )

    def __repr__(self) -> str:
        return get_repr(self)

    @classmethod
    def adsr(
        cls,
        attack_time=0.01,
        decay_time=0.3,
        sustain=0.5,
        release_time=1.0,
        peak=1.0,
        curve=-4.0,
        bias=0.0,
    ) -> "Envelope":
        amplitudes = [x + bias for x in [0, peak, peak * sustain, 0]]
        durations = [attack_time, decay_time, release_time]
        curves = [curve]
        release_node = 2
        return Envelope(
            amplitudes=amplitudes,
            durations=durations,
            curves=curves,
            release_node=release_node,
        )

    @classmethod
    def asr(
        cls, attack_time=0.01, sustain=1.0, release_time=1.0, curve=-4.0
    ) -> "Envelope":
        amplitudes = [0, sustain, 0]
        durations = [attack_time, release_time]
        curves = [curve]
        release_node = 1
        return Envelope(
            amplitudes=amplitudes,
            durations=durations,
            curves=curves,
            release_node=release_node,
        )

    @classmethod
    def from_segments(
        cls,
        initial_amplitude=0,
        segments=None,
        release_node=None,
        loop_node=None,
        offset=None,
    ) -> "Envelope":
        amplitudes = [initial_amplitude]
        durations = []
        curves = []
        for amplitude, duration, curve in segments:
            amplitudes.append(amplitude)
            durations.append(duration)
            curves.append(curve)
        return cls(
            amplitudes=amplitudes,
            durations=durations,
            curves=curves,
            release_node=release_node,
            loop_node=loop_node,
            offset=offset,
        )

    @classmethod
    def percussive(
        cls,
        attack_time: Union[UGenOperable, float] = 0.01,
        release_time: Union[UGenOperable, float] = 1.0,
        amplitude: Union[UGenOperable, float] = 1.0,
        curve: Union[EnvelopeShape, UGenOperable, float, str] = -4.0,
    ) -> "Envelope":
        """
        Make a percussion envelope.

        ::

            >>> from supriya.ugens import Envelope
            >>> envelope = Envelope.percussive()
            >>> envelope
            Envelope(
                curves=(-4.0, -4.0),
                durations=(0.01, 1.0),
            )

        ::

            >>> list(envelope.serialize())
            [<0.0>, <2.0>, <-99.0>, <-99.0>, <1.0>, <0.01>, <5.0>, <-4.0>, <0.0>, <1.0>, <5.0>, <-4.0>]
        """
        amplitudes = [0, amplitude, 0]
        durations = [attack_time, release_time]
        curves = [curve]
        return Envelope(amplitudes=amplitudes, durations=durations, curves=curves)

    @classmethod
    def linen(
        cls, attack_time=0.01, sustain_time=1.0, release_time=1.0, level=1.0, curve=1
    ) -> "Envelope":
        amplitudes = [0, level, level, 0]
        durations = [attack_time, sustain_time, release_time]
        curves = [curve]
        return Envelope(amplitudes=amplitudes, durations=durations, curves=curves)

    def serialize(self, **kwargs) -> UGenVector:
        result: List[Union[UGenOperable, float]] = []
        result.append(self.initial_amplitude)
        result.append(len(self.envelope_segments))
        result.append(-99 if self.release_node is None else self.release_node)
        result.append(-99 if self.loop_node is None else self.loop_node)
        for amplitude, duration, curve in self._envelope_segments:
            result.append(amplitude)
            result.append(duration)
            if isinstance(curve, EnvelopeShape):
                shape = int(curve)
                curve = 0.0
            else:
                shape = 5
            result.append(shape)
            result.append(curve)
        return UGenVector(*result)

    def serialize_interpolated(self) -> UGenVector:
        result: List[Union[UGenOperable, float]] = []
        result.append(self.offset or 0.0)
        result.append(self.initial_amplitude)
        result.append(len(self.envelope_segments))
        result.append(self.duration)
        for amplitude, duration, curve in self._envelope_segments:
            result.append(duration)
            if isinstance(curve, EnvelopeShape):
                shape = int(curve)
                curve = 0.0
            else:
                shape = 5
            result.append(shape)
            result.append(curve)
            result.append(amplitude)
        return UGenVector(*result)

    @classmethod
    def triangle(cls, duration=1.0, amplitude=1.0) -> "Envelope":
        """
        Make a triangle envelope.

        ::

            >>> from supriya.ugens import Envelope
            >>> envelope = Envelope.triangle()
            >>> envelope
            Envelope(
                durations=(0.5, 0.5),
            )

        ::

            >>> list(envelope.serialize())
            [<0.0>, <2.0>, <-99.0>, <-99.0>, <1.0>, <0.5>, <1.0>, <0.0>, <0.0>, <0.5>, <1.0>, <0.0>]
        """
        amplitudes = [0, amplitude, 0]
        duration = duration / 2.0
        durations = [duration, duration]
        return Envelope(amplitudes=amplitudes, durations=durations)

    ### PUBLIC PROPERTIES ###

    @property
    def amplitudes(self) -> Tuple[Union[UGenOperable, float]]:
        return (self.initial_amplitude,) + tuple(_[0] for _ in self.envelope_segments)

    @property
    def curves(self) -> Tuple[Union[EnvelopeShape, UGenOperable, float]]:
        return tuple(_[2] for _ in self.envelope_segments)

    @property
    def duration(self) -> Union[float, UGenOperable]:
        return sum(self.durations)

    @property
    def durations(self) -> Tuple[Union[float, UGenOperable]]:
        return tuple(_[1] for _ in self.envelope_segments)

    @property
    def envelope_segments(self):
        return self._envelope_segments

    @property
    def initial_amplitude(self) -> Union[float, UGenOperable]:
        return self._initial_amplitude

    @property
    def loop_node(self) -> Optional[int]:
        return self._loop_node

    @property
    def offset(self) -> Union[float, UGenOperable]:
        return self._offset

    @property
    def release_node(self) -> Optional[int]:
        return self._release_node


@ugen(kr=True)
class Done(UGen):
    """
    Triggers when `source` sets its `done` flag.

    ::

        >>> source = supriya.ugens.Line.kr()
        >>> done = supriya.ugens.Done.kr(
        ...     source=source,
        ... )
        >>> done
        <Done.kr()[0]>
    """

    source = param()


@ugen(ar=True, kr=True, has_done_flag=True)
class EnvGen(UGen):
    """
    An envelope generator.

    ::

        >>> from supriya.ugens import Envelope, EnvGen
        >>> EnvGen.ar(envelope=Envelope.percussive())
        <EnvGen.ar()[0]>
    """

    gate = param(1.0)
    level_scale = param(1.0)
    level_bias = param(0.0)
    time_scale = param(1.0)
    done_action = param(0.0)
    envelope = param(unexpanded=True)

    @classmethod
    def _new_expanded(
        cls,
        *,
        calculation_rate=None,
        done_action=None,
        envelope=None,
        gate=1.0,
        level_bias=0.0,
        level_scale=1.0,
        time_scale=1.0,
    ):
        if not isinstance(done_action, Parameter):
            done_action = DoneAction.from_expr(done_action)
        return super(EnvGen, cls)._new_expanded(
            calculation_rate=calculation_rate,
            done_action=done_action,
            envelope=(envelope or Envelope()).serialize(),
            gate=gate,
            level_bias=level_bias,
            level_scale=level_scale,
            time_scale=time_scale,
        )


@ugen(kr=True)
class Free(UGen):
    """
    Frees the node at `node_id` when triggered by `trigger`.

    ::

        >>> node_id = 1000
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1.0)
        >>> free = supriya.ugens.Free.kr(
        ...     node_id=node_id,
        ...     trigger=trigger,
        ... )
        >>> free
        <Free.kr()[0]>
    """

    trigger = param(0)
    node_id = param()


@ugen(kr=True)
class FreeSelf(UGen):
    """
    Frees the enclosing synth when triggered by `trigger`.

    ::

        >>> trigger = supriya.ugens.Impulse.kr(frequency=1.0)
        >>> free_self = supriya.ugens.FreeSelf.kr(
        ...     trigger=trigger,
        ... )
        >>> free_self
        <FreeSelf.kr()[0]>
    """

    trigger = param()


@ugen(kr=True)
class FreeSelfWhenDone(UGen):
    """
    Frees the enclosing synth when `source` sets its `done` flag.

    ::

        >>> source = supriya.ugens.Line.kr()
        >>> free_self_when_done = supriya.ugens.FreeSelfWhenDone.kr(
        ...     source=source,
        ... )
        >>> free_self_when_done
        <FreeSelfWhenDone.kr()[0]>
    """

    source = param()


@ugen(kr=True, has_done_flag=True)
class Linen(UGen):
    """
    A simple line generating unit generator.

    ::

        >>> supriya.ugens.Linen.kr()
        <Linen.kr()[0]>
    """

    gate = param(1.0)
    attack_time = param(0.01)
    sustain_level = param(1.0)
    release_time = param(1.0)
    done_action = param(0)


@ugen(kr=True)
class Pause(UGen):
    """
    Pauses the node at `node_id` when triggered by `trigger`.

    ::

        >>> node_id = 1000
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1.0)
        >>> pause = supriya.ugens.Pause.kr(
        ...     node_id=node_id,
        ...     trigger=trigger,
        ... )
        >>> pause
        <Pause.kr()[0]>
    """

    trigger = param()
    node_id = param()


@ugen(kr=True)
class PauseSelf(UGen):
    """
    Pauses the enclosing synth when triggered by `trigger`.

    ::

        >>> trigger = supriya.ugens.Impulse.kr(frequency=1.0)
        >>> pause_self = supriya.ugens.PauseSelf.kr(
        ...     trigger=trigger,
        ... )
        >>> pause_self
        <PauseSelf.kr()[0]>
    """

    trigger = param()


@ugen(kr=True)
class PauseSelfWhenDone(UGen):
    """
    Pauses the enclosing synth when `source` sets its `done` flag.

    ::

        >>> source = supriya.ugens.Line.kr()
        >>> pause_self_when_done = supriya.ugens.PauseSelfWhenDone.kr(
        ...     source=source,
        ... )
        >>> pause_self_when_done
        <PauseSelfWhenDone.kr()[0]>
    """

    source = param()



from IPython.display import Audio, display  # type: ignore

from supriya.ext import websafe_audio
from supriya.io import Player


def load_ipython_extension(ipython):
    ipython.extension_manager.load_extension("uqbar.ext.ipython")
    patch_player()


def patch_player():
    async def patched_call(self):
        output_path, status_code = await self.renderable.__render__(
            **self.render_kwargs,
        )
        if output_path:
            output_path = websafe_audio(output_path)
            display(Audio(filename=str(output_path)))
        return output_path, status_code

    Player.__call__ = patched_call
